

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "rules_callbacks.h"
#include "rules_interface.h"
#include "support.h"

#include "../bozo.h"
#include "../gnksa.h"
#include "../util.h"

#include <gnome.h>

static const gchar*
get_selected_bozo (GtkCList * clist, int * setme_row)
{
	gchar * retval = NULL;
	GList * l;
       
	l = clist->selection;
	if (l != NULL) {
		int row = GPOINTER_TO_INT(l->data);
		if (setme_row != NULL)
			*setme_row = row;
		gtk_clist_get_text (clist, row, 0, &retval);
	}

	return retval;
}

static void
set_selected_bozo (GtkCList * clist, const gchar * name)
{
	pan_lock ();
	if (clist->selection != NULL)
	{
		gint sel_row = GPOINTER_TO_INT(clist->selection->data);
		gtk_clist_set_text (clist, sel_row, 0, name);
	}
	pan_unlock ();
}

void
on_bozo_dialog_realize                 (GtkWidget       *widget,
                                        gpointer         user_data)
{
	GtkCList * clist = GTK_CLIST(lookup_widget (widget, "bozo_clist"));
	GPtrArray * a = bozo_get_list ();
	guint i;

	/* not yet settable through glade */
	gtk_clist_set_sort_column (clist, 0);
	gtk_clist_set_auto_sort (clist, TRUE);

	/* add the bozos to the clist */
	g_message ("got %u bozos", a->len);
	for (i=0; i!=a->len; ++i) {
		gchar * name = (gchar*) g_ptr_array_index (a, i);
		gtk_clist_insert (clist, -1, &name);
	}

	/* cleanup */
	pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
	g_ptr_array_free (a, TRUE);
}


/**
***  Add a bozo
**/

static void
bozo_add_cb (gchar * string, gpointer data)
{
	gtk_clist_insert (GTK_CLIST(data), -1, &string);
}
void
on_bozo_create_button_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkWidget * widget = GTK_WIDGET(button);
	GtkCList * clist = GTK_CLIST(lookup_widget (widget, "bozo_clist"));
	GtkWidget * bozo_dialog = lookup_widget (widget, "bozo_dialog");

	pan_lock ();
	gnome_request_dialog (FALSE,
		              _("Add Bozo in the form\n\"J. Random Bozo\" <bozo@spammesensless.net>"),
			      "",
			      128,
			      bozo_add_cb,
			      clist,
			      GTK_WINDOW(bozo_dialog));
	pan_unlock ();
}

/**
***  Edit a bozo
**/

static void
bozo_rename_cb (gchar * string, gpointer data)
{
	if (is_nonempty_string (string))
		set_selected_bozo (GTK_CLIST(data), string);
}
static void
edit_selected_bozo (GtkCList * clist)
{
	const gchar * bozo = get_selected_bozo (clist, NULL);

	if (bozo != NULL)
	{
		GtkWidget * w = lookup_widget (GTK_WIDGET(clist),"bozo_dialog");

		pan_lock ();
		gnome_request_dialog (FALSE,
		                      _("Edit Bozo's Name"),
				      bozo,
				      128,
				      bozo_rename_cb,
				      clist,
				      GTK_WINDOW(w));
		pan_unlock ();
	}
}

void
on_bozo_edit_button_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
	edit_selected_bozo (GTK_CLIST(
		lookup_widget (GTK_WIDGET(button), "bozo_clist")));
}

void
on_bozo_clist_select_row               (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
					GdkEvent        *event,
					gpointer         user_data)
{
        GdkEventButton * bevent = (GdkEventButton *)event;
	if (bevent!=NULL
		&& bevent->button==1
		&& bevent->type==GDK_2BUTTON_PRESS)
			edit_selected_bozo (clist);

}


/**
***  Delete a bozo
**/

void
on_bozo_delete_button_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkWidget * widget = GTK_WIDGET(button);
	GtkCList * clist = GTK_CLIST(lookup_widget (widget, "bozo_clist"));
	gint row = -1;
	get_selected_bozo (clist, &row);
	if (row != -1)
		gtk_clist_remove (clist, row);
}


/**
***  Close Dialog
**/

void
on_bozo_dialog_clicked                 (GnomeDialog     *gnomedialog,
                                        gint             button,
                                        gpointer         user_data)
{
	GtkWidget * widget = GTK_WIDGET(gnomedialog);
	GtkCList * clist = GTK_CLIST(lookup_widget(widget,"bozo_clist"));

	if (button==0) /* ok */
	{
		gint row;

		/* repopulate the bozo list from the UI */
		bozo_clear ();
		for (row=0; row<clist->rows; ++row)
		{
			gint i;
			gchar * str_nofree = NULL;
			gchar * addr = NULL;
			gchar * real = NULL;
			gtk_clist_get_text (clist, row, 0, &str_nofree);

			i = gnksa_do_check_from (str_nofree, &addr, &real);
			if (i != 0)
			{
				g_message ("Can't parse \"%s\"", str_nofree);
			}
			else
			{
				if (real != NULL)
					gnksa_strip_realname (real);
				bozo_add (addr, real);
			}

			g_free (addr);
			g_free (real);
		}
	}

	gnome_dialog_close (gnomedialog);
}

gint
on_bozo_dialog_close                   (GnomeDialog     *gnomedialog,
                                        gpointer         user_data)
{
	return 0;
}

