/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * dialog-about.c: About dialog
 */

#include <config.h>
#include <gnome.h>

#include "dialogs.h"
#include "globals.h"
#include "gui.h"
#include "util.h"


/**
 * dialog_about:
 *
 * About window to display program information, version number, and
 * copyright.
 **/
void
dialog_about (void)
{
	static GtkWidget *about = NULL;
	const gchar *authors[] = {
		"Matt Eagleson <e.Messiah@superpimp.org> - Project Originator",
		"Tov are Jacobsen <tovj@stud.ntnu.no> - Translations & PR",
		"Charles Kerr <charles@superpimp.org> - Programmer, 0.6 - present",
		"Jason Leach <elerium@superpimp.org> - Programmer, 0.6 - 0.7",
		NULL
	};

        if (about != NULL)
        {
                g_return_if_fail (GTK_WIDGET_REALIZED (about));
                gdk_window_show (about->window);
                gdk_window_raise (about->window);
		return;
        }

	pan_lock ();

	about = gnome_about_new ("Pan", VERSION,
				 _("Copyright (C) 1999, 2000 the Pan Development Group"),
				 authors,
				 _("PAN, a newsreader for GNOME."),
				 PAN_PIXMAPDIR "/pan_druid.xpm");

	gtk_signal_connect (GTK_OBJECT (about), "destroy",
			    GTK_SIGNAL_FUNC (gtk_widget_destroyed),
			    &about);

	if (Pan.window)
		gnome_dialog_set_parent (GNOME_DIALOG (about),
					 GTK_WINDOW (Pan.window));

	gtk_widget_show (about);

	pan_unlock ();
}
