/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <string.h>
#include <unistd.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <glib.h>

#include "debug.h"
#include "gnksa.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PRIVATE
************/

/*****
******
*****/

/*****
******
*****/

static gboolean
is_unquoted_char (int ch)
{
	/* ASCII printable character except !()<>@,;:\".[] */
	return isgraph(ch) && strchr("!()<>@,;:\\\".[]", ch)==NULL;
}

static gboolean
is_quoted_char (int ch)
{
	/* ASCII printable character except "()<>\ */
	return isgraph(ch) && strchr("\"()<>\\", ch)==NULL;
}

static gboolean
is_tag_char (int ch)
{
	/* ascii printable character except !()<>@,;:\"[]/?=> */
	return isgraph(ch) && strchr("!()<>@,;\\\"[]/?=", ch)==NULL;
}

static gboolean
is_code_char (int ch)
{
	/* ascii printable character except ? */
	return isgraph(ch) && ch!='?';
}

static gboolean
is_paren_char (int ch)
{
	/* ascii printable character except ()<>\ */
	return isgraph(ch) && strchr("()<>\\",ch)==NULL;
}

/**
***
**/

static gboolean
read_space (const char * start, const char ** end)
{
	/* 1*( <HT (ASCII 9)> / <blank (ASCII 32)> */
	const gchar * pch = start;
	while (*pch=='\t' || *pch==' ') ++pch;
	if (pch != start) {
		*end = pch;
		return TRUE;
	}
	return FALSE;
	
}

static gboolean
read_codes (const char * pch, const char ** end)
{
	if (!is_code_char(*pch)) return FALSE;
	while (is_code_char(*pch)) ++pch;
	*end = pch;
	return TRUE;
}

static gboolean
read_encoding (const char * pch, const char ** end)
{
	if (!is_tag_char(*pch)) return FALSE;
	while (is_tag_char(*pch)) ++pch;
	*end = pch;
	return TRUE;
}

static gboolean
read_charset (const char * pch, const char ** end)
{
	return read_encoding (pch, end);
}

static gboolean
read_encoded_word (const char * pch, const char ** end)
{
	/* "=?" charset "?" encoding "?" codes "?=" */
	if (pch[0]!='=' || pch[1]!='?') return FALSE;
	pch += 2;
	if (!read_charset (pch, &pch)) return FALSE;
	if (*pch != '?') return FALSE;
	++pch;
	if (!read_encoding (pch, &pch)) return FALSE;
	if (*pch != '?') return FALSE;
	++pch;
	if (!read_codes (pch, &pch)) return FALSE;
	if (pch[0]!='?' || pch[1]!='=') return FALSE;
	*end = pch + 2;
	return TRUE;
}

static gboolean
read_unquoted_word (const char * pch, const char ** end)
{
	/* 1*unquoted-char */
	if (!is_unquoted_char(*pch)) return FALSE;
	while (is_unquoted_char((int)*pch)) ++pch;
	*end = pch;
	return TRUE;
}

static gboolean
read_quoted_word (const char * pch, const char ** end)
{
	/* quote 1*(quoted-char / space) quote */
	if (*pch!='"') return FALSE;
	++pch;
	for (;;) {
		if (read_space(pch,&pch))
			continue;
		if (is_quoted_char ((int)*pch)) {
			++pch;
			continue;
		}
		break;
	}
	if (*pch!='"') return FALSE;
	++pch;
	*end = pch;
	return TRUE;
}

static gboolean
read_plain_word (const char * pch, const char ** end)
{
	/* unquoted-word / quoted-word / encoded-word */
	if (read_unquoted_word(pch,end)) return TRUE;
	if (read_quoted_word(pch,end)) return TRUE;
	if (read_encoded_word(pch,end)) return TRUE;
	return FALSE;
}

static gboolean
read_plain_phrase (const char * pch, const char ** end)
{
	/* plain-word *(space plain-word) */
	const char * tmp = NULL;
	if (!read_plain_word(pch,&pch))
		return FALSE;
	for (;;) {
		tmp = pch;
		if (!read_space(pch,&pch)) break;
		if (!read_plain_word(pch,&pch)) break;
	}
	*end = tmp;
	return TRUE;
}

static gboolean
read_paren_char (const char * pch, const char ** end)
{
	if (!is_paren_char((int)*pch))
		return FALSE;
	*end = pch + 1;
	return TRUE;
}


static gboolean
read_paren_phrase (const char * pch, const char ** end)
{
	/* 1* (paren-char / space / encoded-word */

	if (!read_paren_char(pch,&pch)
		&& !read_space(pch,&pch)
		&& !read_encoded_word(pch,&pch))
			return FALSE;

	for (;;)
	{
		if (!read_paren_char(pch,&pch)
			&& !read_space(pch,&pch)
			&& !read_encoded_word(pch,&pch))
				break;
	}

	*end = pch;
	return TRUE;
}


/*****
******
*****/

/*****
******
*****/

static int
gnksa_split_from (const char * from,
                  char ** alloc_and_setme_address,
		  char ** alloc_and_setme_realname,
		  int * addrtype)
{
	char * begin;
	char * end;
	char * work;

	*alloc_and_setme_address = NULL;
	*alloc_and_setme_realname = NULL;

	work = g_strdup (from);
	g_strstrip (work);

	/* empty string */
	if (!*work)
	{
		*addrtype = GNKSA_ADDRTYPE_OLDSTYLE;
		g_free (work);
		return GNKSA_LPAREN_MISSING;
	}

	end = &work [ strlen(work) - 1 ];
	debug2 (DEBUG_SOCKET_OUTPUT, "work: [%s], end: [%s]", work, end);
	if (*end == '>') /* Charles Kerr <charles@superpimp.org> */
	{
		const char * tmp;
		*addrtype = GNKSA_ADDRTYPE_ROUTE;

		/* get address part */
		begin = strrchr (work, '<');
		if (begin == NULL) {
			g_free (work);
			return GNKSA_LANGLE_MISSING;
		}

		/* copy route address from inside the <> brackets */
		*alloc_and_setme_address = g_strdup (begin+1);
		*strchr(*alloc_and_setme_address, '>') = '\0';

		/* From: [plain-phrase space] "<" address ">" */
		*begin = '\0';
		tmp = work;
		if ((*tmp) && (!read_plain_phrase(tmp,&tmp)||!read_space(tmp,&tmp))) {
			g_free (work);
			return GNKSA_ILLEGAL_PLAIN_PHRASE;
		}

		/* get realname part */
		*alloc_and_setme_realname = g_strdup (work);
	}
	else /* charles@superpimp.org (Charles Kerr) */
	{
		*addrtype = GNKSA_ADDRTYPE_OLDSTYLE;

		/* get address part  */
		begin = work;
		for (end=begin; *end && !isspace((int)*end); ) ++end;
		*alloc_and_setme_address = g_strndup (begin, end-begin);
		debug1 (DEBUG_SOCKET_OUTPUT, "tmp address: [%s]", begin);

		/* get realname part */
		begin = end;
		if (*begin) ++begin;
		g_strstrip (begin);
		debug1 (DEBUG_SOCKET_OUTPUT, "tmp real: [%s]", begin);
		if (!*begin)
		{
			*alloc_and_setme_realname = NULL;
		}
		else
		{
			const int len = strlen (begin);
			const char * tmp = NULL;
			if (*begin != '(') {
				g_free (work);
				return GNKSA_LPAREN_MISSING;
			}
			else if (begin[len-1]!=')') {
				g_free (work);
				return GNKSA_RPAREN_MISSING;
			}
			begin[len-1] = '\0';
			if (!read_paren_phrase(begin+1,&tmp) || tmp==NULL) {
				g_free (work);
				return GNKSA_ILLEGAL_PAREN_PHRASE;
			}
			begin[len-1] = ')';
			*alloc_and_setme_realname = g_strdup(begin);
		}
	}

	g_free (work);
	return GNKSA_OK;
}

/*****
******
*****/

/**
 *  "[244.14.124.12]"
 *  "244.12.14.12"
 */
static int
gnksa_check_domain_literal (const char* domain)
{
	int i;
	int x[4];
	gboolean need_closing_brace;
	const char * pch;

	/* parse domain literal into ip number */

	pch = domain;
	need_closing_brace = *pch == '[';
	if (need_closing_brace) ++pch; /* move past open brace */

	/* %u.%u.%u.%u */
	for (i=0; i<4; ++i) {
		char * end = NULL;
		x[i] = strtoul (pch, &end, 10);
		if (end == pch)
			return GNKSA_BAD_DOMAIN_LITERAL;
		if (x[i]<0 || x[i]>255)
			return GNKSA_BAD_DOMAIN_LITERAL;
		if (i!=3) {
			if (*end != '.')
				return GNKSA_BAD_DOMAIN_LITERAL;
			++end;
		}
		pch = end;
	}

	if (need_closing_brace && *pch!=']')
		return GNKSA_BAD_DOMAIN_LITERAL;

	return GNKSA_OK;
}

/*****
******
*****/

static int
gnksa_check_domain (const char * domain)
{
	int i;
	int label_qty;
	char ** labels;
	const char * top;

	/* check for domain literal */
	if (*domain == '[')
		return gnksa_check_domain_literal (domain);

	/* check for empty */
	if (!is_nonempty_string (domain)
		|| *domain=='.'
		|| domain[strlen(domain)-1]=='.'
		|| strstr(domain,"..")!=NULL)
		return GNKSA_ZERO_LENGTH_LABEL;

	/* count the labels */
	labels = g_strsplit (domain, ".", -1);
	for (label_qty=0; labels[label_qty]!=NULL; ) {
		++label_qty;
	}

	/* make sure we have more than one label in the domain */
	if (label_qty < 2) {
		g_strfreev (labels);
		return GNKSA_SINGLE_DOMAIN;
	}

	/* check for illegal labels */
	for (i=0; i<label_qty-1; ++i) {
		const char * label = labels[i];
		if (strlen(label) > 63) {
			g_strfreev (labels);
			return GNKSA_ILLEGAL_LABEL_LENGTH;
		}
		if (label[0]=='-' || label[strlen(label)-1]=='-') {
			g_strfreev (labels);
			return GNKSA_ILLEGAL_LABEL_HYPHEN;
		}
	}

	/* check existence of the toplevel domain */
	top = labels[label_qty-1];
	debug1 (DEBUG_SOCKET_OUTPUT, "top level domain: [%s]", top);
	switch ((int)strlen(top))
	{
		case 1:
			if (isdigit((int)top[0]))
			{
				g_strfreev (labels);
				return gnksa_check_domain_literal (domain);
			}

			/* single-letter TLDs dont exist */
			g_strfreev (labels);
			return GNKSA_ILLEGAL_DOMAIN;
			break;

		case 2:
			if (isdigit((int)top[0])
				|| isdigit((int)top[1]))
			{
				g_strfreev (labels);
				return gnksa_check_domain_literal (domain);
			}
			break;
		case 3:
			if (isdigit((int)top[0])
				|| isdigit((int)top[2])
				|| isdigit((int)top[3]))
			{
				g_strfreev (labels);
				return gnksa_check_domain_literal (domain);
			}
			break;
		default:
			break;
	}

	g_strfreev (labels);
	return GNKSA_OK;
}

/*****
******
*****/

static int
gnksa_check_localpart (const char * localpart)
{
	char ** words;
	int i;

	/* make sure it's not empty... */
	if (!is_nonempty_string(localpart))
		return GNKSA_LOCALPART_MISSING;

	/* break localpart up into its unquoted words */
	words = g_strsplit (localpart, ".", -1);
	for (i=0; words[i]!=NULL; ++i)
	{
		const char * word = words[i];
		if (!*word) {
			g_strfreev (words);
			return GNKSA_ZERO_LENGTH_LOCAL_WORD;
		}
		for (; *word; ++word) {
			if (!is_unquoted_char(*word)) {
				g_strfreev (words);
				return GNKSA_INVALID_LOCALPART;
			}
		}
	}

	g_strfreev (words);
	return GNKSA_OK;
}

/*****
******
*****/

/*****
******
*****/

/*****
******
*****/

/************
*************  PUBLIC
************/

/*****
******
*****/

GPtrArray*
gnksa_split_addresses (const char * from)
{
	GPtrArray * a = g_ptr_array_new ();
	gboolean quoted = FALSE;
	const gchar * pch = from;
	const gchar * start = from;

	for (pch=start=from; is_nonempty_string(pch); )
	{
		while (*pch && !strchr("\",;:",*pch))
			++pch;

		if (*pch=='"') /* toggle quoted */
			quoted = !quoted;
		else if (!quoted) /* delimiter */
		{
			gchar * s = g_strndup (start, pch-start);
			g_strstrip (s);
			if (is_nonempty_string(s))
				g_ptr_array_add (a, s);
			else /* too small; throw it back */
				g_free (s);
			if (*pch)
				start = pch + 1;
		}

		if (*pch)
			++pch;
	}

	return a;
}

/*****
******
*****/

gint
gnksa_check_address (const gchar * address)
{
	gint retval = GNKSA_OK;
	gchar * tmp;
	gchar * begin;

	/* get rid of vacuous case */
	if (!is_nonempty_string(address))
		return GNKSA_LOCALPART_MISSING;

	/* check the address */
	tmp = g_strdup (address);
	begin = strrchr (address, '@');
	if (begin==NULL)
	{
		if (retval == GNKSA_OK)
			retval = GNKSA_INVALID_DOMAIN;
	}
	else
	{
		/* temporarily split string between username & fqdn*/
		*begin++ = '\0';
		g_strdown (begin);

		/* check the domain */
		if (retval == GNKSA_OK)
			retval = gnksa_check_domain (begin);

		/* check the localpart */
		if (retval == GNKSA_OK)
			retval = gnksa_check_localpart (address);

		/* restore separator */
		*--begin = '@';
	}

	g_free (tmp);
	return retval;
}

/*****
******
*****/

int
gnksa_do_check_from (const char   * from,
                     char        ** alloc_and_setme_address,
                     char        ** alloc_and_setme_realname)
{
	int addrtype = 0;
	int retval;
	char * address;
	char * realname;
	debug1 (DEBUG_SOCKET_OUTPUT, "From: [%s]", from);

	/* split from */
	retval = gnksa_split_from (from,
	                           alloc_and_setme_address,
		         	   alloc_and_setme_realname, 
			           &addrtype);

	address = *alloc_and_setme_address;
	realname = *alloc_and_setme_realname;
	debug1 (DEBUG_SOCKET_OUTPUT, "Name: [%s]", (realname?realname:"NULL"));
	debug1 (DEBUG_SOCKET_OUTPUT, "Address: [%s]", (address?address:"NULL"));

	/* check address */
	if (address != NULL) {
		if (retval == GNKSA_OK)
			retval = gnksa_check_address (address);
	}

	debug1 (DEBUG_SOCKET_OUTPUT, "GNKSA [%d]", retval);
	return retval;
}

/*****
******
*****/

void
gnksa_strip_realname (gchar * realname)
{
	gchar * pch;

	g_return_if_fail (is_nonempty_string(realname));

	g_strstrip (realname);
	if (*realname=='(') {
		size_t len = strlen (realname+1);
		g_memmove (realname, realname+1, len+1);
		pch = realname + len - 1;
		if (*pch == ')')
			*pch = '\0';
	}

	while ((pch = strchr(realname,'"')) != NULL)
		g_memmove (pch, pch+1, strlen(pch+1)+1);
}

/*****
******
*****/

int
gnksa_check_from (const char* from)
{
	int retval;
	char * address;
	char * realname;

	address = realname = NULL;
	retval = gnksa_do_check_from (from, &address, &realname);
	g_free (address);
	g_free (realname);

	return retval;
}

/*****
******
*****/

/* see son-of-1036 5.3 */

int
gnksa_check_message_id (const char* message_id)
{
	int val;
	gchar * tmp;
	gchar * pch;

	/* make a temporary copy that we can poke */
       	tmp = g_strdup (message_id);
	g_strstrip (tmp);

	/* make sure it's <= 250 octets (son of gnksa 1036 5.3) */
	if (strlen(tmp) > 250) {
		g_free (tmp);
		return GNKSA_ILLEGAL_LABEL_LENGTH;
	}

	/* make sure the message-id is wrapped in < > */
	if (*tmp!='<') {
		g_free (tmp);
		return GNKSA_LANGLE_MISSING;
	}
	if (tmp[strlen(tmp)-1]!='>') {
		g_free (tmp);
		return GNKSA_RANGLE_MISSING;
	}

	/* find the '@' separator */
	pch = strrchr (tmp, '@');
	if (pch == NULL) {
		g_free (tmp);
		return GNKSA_ATSIGN_MISSING;
	}

	/* check the domain name */
	++pch;
	*strrchr(pch,'>') = '\0';
	val = gnksa_check_domain (pch);
	if (val != GNKSA_OK) {
		g_free (tmp);
		return val;
	}

	/* check the local-part */
	--pch;
	*pch = '\0';
	if (!g_strcasecmp (tmp+1, "postmaster")) { /* son-of-1036 */
		g_free (tmp);
		return GNKSA_INVALID_LOCALPART;
	}
	val = gnksa_check_localpart (tmp+1);
	if (val != GNKSA_OK) {
		g_free (tmp);
		return val;
	}

	g_free (tmp);
	return GNKSA_OK;
}

/*****
******
*****/


const char *
gnksa_status_text (int status)
{
	switch (status)
	{
	case GNKSA_OK: return _("Okay");
	case GNKSA_INTERNAL_ERROR: return _("Internal Error");

	case GNKSA_LANGLE_MISSING: return _("'<' Missing");
	case GNKSA_RANGLE_MISSING: return _("'>' Missing");
	case GNKSA_LPAREN_MISSING: return _("'(' Missing");
	case GNKSA_RPAREN_MISSING: return _("')' Missing");
	case GNKSA_ATSIGN_MISSING: return _("'@' Missing");

	case GNKSA_SINGLE_DOMAIN: return _("Single Domain");
	case GNKSA_INVALID_DOMAIN: return _("Invalid Domain");
	case GNKSA_ILLEGAL_DOMAIN: return _("Illegal Domain");
	case GNKSA_UNKNOWN_DOMAIN: return _("Unknown Domain");
	case GNKSA_INVALID_FQDN_CHAR: return _("Invalid char in domain");
	case GNKSA_ZERO_LENGTH_LABEL: return _("Zero Length Label");
	case GNKSA_ILLEGAL_LABEL_LENGTH: return _("Illegal label length");
	case GNKSA_ILLEGAL_LABEL_HYPHEN: return _("Hypen in label");
	case GNKSA_ILLEGAL_LABEL_BEGNUM: return _("fully-qualified domain names must not begin with a digit.");
	case GNKSA_BAD_DOMAIN_LITERAL: return _("Bad domain literal");
	case GNKSA_LOCAL_DOMAIN_LITERAL: return _("Local domain literal");
	case GNKSA_RBRACKET_MISSING: return _("Right bracket missing");

	case GNKSA_LOCALPART_MISSING: return _("Local part missing");
	case GNKSA_INVALID_LOCALPART: return _("Invalid localpart");
	case GNKSA_ZERO_LENGTH_LOCAL_WORD: return _("Zero length local word");

	case GNKSA_ILLEGAL_UNQUOTED_CHAR:return _("Illegal unquoted character");
	case GNKSA_ILLEGAL_QUOTED_CHAR: return _("Illegal quoted character");
	case GNKSA_ILLEGAL_ENCODED_CHAR: return _("Illegal encoded character");
	case GNKSA_BAD_ENCODE_SYNTAX: return _("Bad encode syntax");
	case GNKSA_ILLEGAL_PAREN_PHRASE: return _("Illegal parenthesis phrase");
	case GNKSA_ILLEGAL_PAREN_CHAR: return _("Illegal parenthesis");
	case GNKSA_INVALID_REALNAME: return _("Invalid real name");
	case GNKSA_ILLEGAL_PLAIN_PHRASE: return _("Illegal plain phrase");
	}

	pan_warn_if_reached ();
	return _("Internal Error");
};

/*****
******
*****/

static char*
remove_broken_message_ids_from_references (const char * references)
{
	gchar * pch;
	gchar * tmp;
	GString * fixed;

	/* sanity checks */
	g_return_val_if_fail (is_nonempty_string(references), NULL);

	/* remove broken message-ids
	   FIXME: this tokenizing works, but it's hard to read. */
       	fixed = g_string_sized_new (1024);
       	tmp = g_strdup (references);
	pch = tmp;
	for (;;)
	{
		gchar tmp_end_val;
		gchar * begin;
		gchar * end;

		/* skip to the beginning of a message id */
		while (*pch!='<')
			++pch;
		begin = pch;
		++pch;
		end = strpbrk (pch, "<> ");
		if (end==NULL)
			end = pch + strlen(pch);
		else if (*end == '>')
			++end;

		/* zero-terminate the message-id to check it for validity */
		tmp_end_val = *end;
		*end = '\0';
		if (gnksa_check_message_id(begin)==GNKSA_OK) {
			g_string_append (fixed, begin);
			g_string_append_c (fixed, ' ');
		}
		*end = tmp_end_val;
		pch = end;

		/* if that was the last token, then we're done */
		if (tmp_end_val == '\0')
			break;

	}
	pan_g_string_strstrip (fixed); /* remove trailing space  */

	/* cleanup */
	pch = fixed->str;
	g_free (tmp);
	g_string_free (fixed, FALSE);

	return pch;
}

/**
 * Try to trim references down to an acceptable length for the NNTP server,
 * @param untrimmed "references" string
 * @return newly-allocated trimmed "references" string
 */
static gchar*
gnksa_trim_references_to_len (const gchar* refs, int cutoff)
{
	char * fixed = remove_broken_message_ids_from_references (refs);
	char * retval;
	const int fixed_len = strlen (fixed);

	if (fixed_len < cutoff) /* this is too easy. :) */
	{
		retval = g_strdup(fixed);
	}
	else /* plow through them */
	{
		int leader_len;
		int len_left;
		const char * str_end = fixed + fixed_len;
		const char * pch;

		/* isolate first id, which we must always keep */
	       	pch = fixed;
		while (*pch!='<') ++pch; /* find beginning of first id */
		while (*pch!='>') ++pch; /* find end of first id */
		pch += 2; /* keep the "> " */
		leader_len = pch - fixed;


		/* trim out 2..n until we fit.  This implicitly lets later
		   ids make the final cut, which is in line with gnksa's
		   requirement that the final three ids not be trimmed out. */
		len_left = cutoff - leader_len;
		while (str_end-pch > len_left) {
			++pch; /* skip until we fit */
		}
		while (*pch && *pch!='<' ) ++pch; /* can't include half an id */

		/* generate return string */
		retval = g_malloc (leader_len + (str_end-pch) + 1);
		g_memmove (retval, fixed, leader_len);
		g_memmove (retval+leader_len, pch, (str_end-pch)+1); /*+1 for end*/
	}

	/* return */
	g_free (fixed);
	pan_warn_if_fail (strlen(retval) <= cutoff);
	return retval;
}

/**
 * thus spake son-of-1036: "the most popular method of generating local parts
 * is to use the date and time, plus some way of distinguishing between
 * simultaneous postings on the same host (e.g., a process number), and encode
 * them in a suitably-reduced alphabet.
 */
gchar*
gnksa_generate_message_id (const gchar * author_addr)
{
	char * tmp;
	GString * msgid = g_string_sized_new (256);

	/* start with '<' */
	g_string_append_c (msgid, '<');

	/* add local part to message-id */
	{
		char buf[32];
		const time_t now = time(NULL);
		struct tm local;
		localtime_r (&now, &local);
		strftime (buf, sizeof(buf), "%Y%m%d.%H%M%S", &local);
		g_string_append (msgid, buf);
		g_string_sprintfa (msgid, ".%d.%lu", rand(), (gulong)getpid());
	}

	/* delimit */
	g_string_append_c (msgid, '@');

	/* add fqdn to message-id */
	{
		gchar * host = get_host_name ();
		gchar * fqdn = get_fqdn (host);
		if (is_nonempty_string(fqdn))
		{
			g_string_append (msgid, fqdn);
		}
		else if (is_nonempty_string(author_addr))
		{
			const gchar * pch = strchr (author_addr, '@');
			if (pch != NULL)
				g_string_append (msgid, pch+1);
		}
		else g_string_append (msgid, "pancantresolveyourdomain.net");
		g_free (host);
		g_free (fqdn);
	}

	/* end with '>' */
	g_string_append_c (msgid, '>');

	/* return */
	tmp = msgid->str;
	g_string_free (msgid, FALSE);
	return tmp;
}

gchar*
gnksa_trim_references (const gchar* refs)
{
	const int GNKSA_CUTOFF = 998; /* gnksa rule 7 */
	return gnksa_trim_references_to_len (refs, GNKSA_CUTOFF);
}

gchar*
gnksa_generate_references (const gchar * references,
                           const gchar * message_id)
{
	gchar * retval = NULL;

	if (is_nonempty_string (message_id))
	{
		/* build a full references string */
		GString * str = g_string_sized_new (1024);
		if (is_nonempty_string(references)) {
			g_string_append (str, references);
			g_string_append_c (str, ' ');
		}
		g_string_append (str, message_id);

		/* make sure they fit gnksa guidelines */
		retval = gnksa_trim_references (str->str);
		g_string_free (str, TRUE);
	}

	return retval;
}


/*****
******
*****/

typedef struct 
{
	int expected_retval;
	const char * address;
}
AddressCheck;

typedef struct
{
	int len;
	const char * in;
	const char * expected_out;
}
ReferencesCheck;

void
gnksa_regression_tests (void)
{
	ReferencesCheck references [] = {
		{
			998, /*  gnksa cutoff */

			"<gnksa_pan-0.8.0_001@lull.org>"
			" <gnksa_pan-0.8.0_002@lull.org> <gnksa_pan-0.8.0_003@lull.org>"
			" <gnksa_pan-0.8.0_004@lull.org> <gnksa_pan-0.8.0_005@lull.org>"
			" <gnksa_pan-0.8.0_006@lull.org> <gnksa_pan-0.8.0_007@lull.org>"
			" <gnksa_pan-0.8.0_008@lull.org> <gnksa_pan-0.8.0_009@lull.org>"
			" <gnksa_pan-0.8.0_010@lull.org> <gnksa_pan-0.8.0_011@lull.org>"
			" <gnksa_pan-0.8.0_012@lull.org> <gnksa_pan-0.8.0_013@lull.org>"
			" <gnksa_pan-0.8.0_014@lull.org> <gnksa_pan-0.8.0_015@lull.org>"
			" <gnksa_pan-0.8.0_016@lull.org> <gnksa_pan-0.8.0_017@lull.org>"
			" <gnksa_pan-0.8.0_018@lull.org> <gnksa_pan-0.8.0_019@lull.org>"
			" <gnksa_pan-0.8.0_020@lull.org> <gnksa_pan-0.8.0_021@lull.org>"
			" <gnksa_pan-0.8.0_022@lull.org> <gnksa_pan-0.8.0_023@lull.org>"
			" <gnksa_pan-0.8.0_024@lull.org> <gnksa_pan-0.8.0_025@lull.org>"
			" <gnksa_pan-0.8.0_026@lull.org> <gnksa_pan-0.8.0_027@lull.org>"
			" <gnksa_pan-0.8.0_028@lull.org> <gnksa_pan-0.8.0_029@lull.org>"
			" <gnksa_pan-0.8.0_030@lull.org> <gnksa_pan-0.8.0_031@lull.org>"
	  		" <gnksa_pan-0.8.0_035.12345@lull.org>",

			"<gnksa_pan-0.8.0_001@lull.org>"
			" <gnksa_pan-0.8.0_002@lull.org> <gnksa_pan-0.8.0_003@lull.org>"
			" <gnksa_pan-0.8.0_004@lull.org> <gnksa_pan-0.8.0_005@lull.org>"
			" <gnksa_pan-0.8.0_006@lull.org> <gnksa_pan-0.8.0_007@lull.org>"
			" <gnksa_pan-0.8.0_008@lull.org> <gnksa_pan-0.8.0_009@lull.org>"
			" <gnksa_pan-0.8.0_010@lull.org> <gnksa_pan-0.8.0_011@lull.org>"
			" <gnksa_pan-0.8.0_012@lull.org> <gnksa_pan-0.8.0_013@lull.org>"
			" <gnksa_pan-0.8.0_014@lull.org> <gnksa_pan-0.8.0_015@lull.org>"
			" <gnksa_pan-0.8.0_016@lull.org> <gnksa_pan-0.8.0_017@lull.org>"
			" <gnksa_pan-0.8.0_018@lull.org> <gnksa_pan-0.8.0_019@lull.org>"
			" <gnksa_pan-0.8.0_020@lull.org> <gnksa_pan-0.8.0_021@lull.org>"
			" <gnksa_pan-0.8.0_022@lull.org> <gnksa_pan-0.8.0_023@lull.org>"
			" <gnksa_pan-0.8.0_024@lull.org> <gnksa_pan-0.8.0_025@lull.org>"
			" <gnksa_pan-0.8.0_026@lull.org> <gnksa_pan-0.8.0_027@lull.org>"
			" <gnksa_pan-0.8.0_028@lull.org> <gnksa_pan-0.8.0_029@lull.org>"
			" <gnksa_pan-0.8.0_030@lull.org> <gnksa_pan-0.8.0_031@lull.org>"
	  		" <gnksa_pan-0.8.0_035.12345@lull.org>"
		},
		{
			998, /*  gnksa cutoff */

			"<gnksa_pan-0.8.0_001@lull.org>"
			" <gnksa_pan-0.8.1_001@lull.org>"
			" <gnksa_pan-0.8.0_002@lull.org> <gnksa_pan-0.8.0_003@lull.org>"
			" <gnksa_pan-0.8.0_004@lull.org> <gnksa_pan-0.8.0_005@lull.org>"
			" <gnksa_pan-0.8.0_006@lull.org> <gnksa_pan-0.8.0_007@lull.org>"
			" <gnksa_pan-0.8.0_008@lull.org> <gnksa_pan-0.8.0_009@lull.org>"
			" <gnksa_pan-0.8.0_010@lull.org> <gnksa_pan-0.8.0_011@lull.org>"
			" <gnksa_pan-0.8.0_012@lull.org> <gnksa_pan-0.8.0_013@lull.org>"
			" <gnksa_pan-0.8.0_014@lull.org> <gnksa_pan-0.8.0_015@lull.org>"
			" <gnksa_pan-0.8.0_016@lull.org> <gnksa_pan-0.8.0_017@lull.org>"
			" <gnksa_pan-0.8.0_018@lull.org> <gnksa_pan-0.8.0_019@lull.org>"
			" <gnksa_pan-0.8.0_020@lull.org> <gnksa_pan-0.8.0_021@lull.org>"
			" <gnksa_pan-0.8.0_022@lull.org> <gnksa_pan-0.8.0_023@lull.org>"
			" <gnksa_pan-0.8.0_024@lull.org> <gnksa_pan-0.8.0_025@lull.org>"
			" <gnksa_pan-0.8.0_026@lull.org> <gnksa_pan-0.8.0_027@lull.org>"
			" <gnksa_pan-0.8.0_028@lull.org> <gnksa_pan-0.8.0_029@lull.org>"
			" <gnksa_pan-0.8.0_030@lull.org> <gnksa_pan-0.8.0_031@lull.org>"
	  		" <gnksa_pan-0.8.0_035.12345@lull.org>",

			"<gnksa_pan-0.8.0_001@lull.org>"
			" <gnksa_pan-0.8.0_002@lull.org> <gnksa_pan-0.8.0_003@lull.org>"
			" <gnksa_pan-0.8.0_004@lull.org> <gnksa_pan-0.8.0_005@lull.org>"
			" <gnksa_pan-0.8.0_006@lull.org> <gnksa_pan-0.8.0_007@lull.org>"
			" <gnksa_pan-0.8.0_008@lull.org> <gnksa_pan-0.8.0_009@lull.org>"
			" <gnksa_pan-0.8.0_010@lull.org> <gnksa_pan-0.8.0_011@lull.org>"
			" <gnksa_pan-0.8.0_012@lull.org> <gnksa_pan-0.8.0_013@lull.org>"
			" <gnksa_pan-0.8.0_014@lull.org> <gnksa_pan-0.8.0_015@lull.org>"
			" <gnksa_pan-0.8.0_016@lull.org> <gnksa_pan-0.8.0_017@lull.org>"
			" <gnksa_pan-0.8.0_018@lull.org> <gnksa_pan-0.8.0_019@lull.org>"
			" <gnksa_pan-0.8.0_020@lull.org> <gnksa_pan-0.8.0_021@lull.org>"
			" <gnksa_pan-0.8.0_022@lull.org> <gnksa_pan-0.8.0_023@lull.org>"
			" <gnksa_pan-0.8.0_024@lull.org> <gnksa_pan-0.8.0_025@lull.org>"
			" <gnksa_pan-0.8.0_026@lull.org> <gnksa_pan-0.8.0_027@lull.org>"
			" <gnksa_pan-0.8.0_028@lull.org> <gnksa_pan-0.8.0_029@lull.org>"
			" <gnksa_pan-0.8.0_030@lull.org> <gnksa_pan-0.8.0_031@lull.org>"
	  		" <gnksa_pan-0.8.0_035.12345@lull.org>"
		},
		{
			1024,

			"<gnksa.01@lull.org> <lull.org> <gnksa.03> <@lull.org>"
			" <gnksa.05@> <gnksa.06@lull.org> <>"
			" <gnksa.08@lull.org> <gnksa.09@lull.org"
			" <gnksa.10@lull.org> gnksa.11@lull.org>"
			" <gnksa.12@@lull.org> <gnksa.13@lull.org>.14@lull.org>"
			" gnksa.15@lull.org <gnksa.16@lull.org>",

			"<gnksa.01@lull.org> <gnksa.06@lull.org>"
			" <gnksa.08@lull.org> <gnksa.10@lull.org>"
			" <gnksa.13@lull.org>"
			" <gnksa.16@lull.org>"
		},
		{
			1024,

			"<a@b.a> <d@f.uk> <postmaster@l.uk> <n@o.uk> <@bar.uk>"
			" <foo@bar.com><baz@xyzzy.org> <foo@> <foo@"
			" <blah@trala> <blah@trala.org>",

			"<d@f.uk> <n@o.uk> <foo@bar.com>"
			" <baz@xyzzy.org> <blah@trala.org>"
		},
		{
			36,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <k@l.uk> <n@o.uk> <q@r.uk>"
		},
		{
			35,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <k@l.uk> <n@o.uk> <q@r.uk>"
		},
		{
			34,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <n@o.uk> <q@r.uk>"
		},
		{
			32,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <n@o.uk> <q@r.uk>",
		},
		{
			27,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <n@o.uk> <q@r.uk>",
		},
		{
			26,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <n@o.uk> <q@r.uk>",
		},
		{
			25,
			"<a@b.uk> <d@f.uk> <h@i.uk> <k@l.uk> <n@o.uk> <q@r.uk>",
			"<a@b.uk> <q@r.uk>",
		},
	};

	AddressCheck addresses [] = {
	{ GNKSA_ILLEGAL_PLAIN_PHRASE, "Charles S. Kerr <charles@foo.com>" },
	{ GNKSA_OK, "\"Charles S. Kerr\" <charles@foo.com>" },
	{ GNKSA_OK, "charles kerr <charles@superpimp.org>" },
	{ GNKSA_OK, "Charles \"Pan Programmer\" Kerr <charles@superpimp.org>" },
	{ GNKSA_OK, "Charles \"Likes, to, put, commas, in, quoted, strings\" Kerr <charles@superpimp.org>" },
	{ GNKSA_OK, "\"Charles Kerr, Pan Programmer\" <charles@superpimp.org>" },
	{ GNKSA_ILLEGAL_PLAIN_PHRASE, "Charles Kerr, Pan Programmer <charles@superpimp.org>" },
	{ GNKSA_INVALID_DOMAIN, "charles kerr <charles>" },
	{ GNKSA_OK, "charles@superpimp.org" },
	{ GNKSA_OK, "charles@superpimp.org (Charles Kerr)" },
	{ GNKSA_SINGLE_DOMAIN, "Charles Kerr <charles@org>" },
	{ GNKSA_SINGLE_DOMAIN, "Charles Kerr <@org>" },
	{ GNKSA_OK, "Charles Kerr <charles@[127.0.0.1]>" },
	{ GNKSA_BAD_DOMAIN_LITERAL, "Charles Kerr <charles@[127..0.1]>" },
	{ GNKSA_BAD_DOMAIN_LITERAL, "Charles Kerr <charles@[127...1]>" },
	{ GNKSA_BAD_DOMAIN_LITERAL, "Charles Kerr <charles@[127.0.0.]>" },
	{ GNKSA_ZERO_LENGTH_LABEL, "<charles@pimp.org.>" },
	{ GNKSA_ILLEGAL_LABEL_HYPHEN, "<charles@pimp-.org>" },
	{ GNKSA_OK, "<charles@pimp.org>" },
	{ GNKSA_OK, "Charles <charles@pimp.asf.fa>" },
	{ GNKSA_OK, "Charles <charles@pimp.asf.uk>" },
	{ GNKSA_ZERO_LENGTH_LABEL, "<charles@>" },
	{ GNKSA_LOCALPART_MISSING, "<@pimp.org>" },
	{ GNKSA_OK, "Charles Kerr <charles@skywalker.ecn.ou.edu>" },
	};
	int qty = sizeof(addresses) / sizeof(addresses[0]);
	int i;
	for (i=0; i!=qty; ++i) {
		int retval = gnksa_check_from (addresses[i].address);
		const char * pass_msg =
			retval==addresses[i].expected_retval ? "PASS" : "FAIL";
		printf ("%s address test #%d (address [%s] got [%d])\n",
			pass_msg, i+1, addresses[i].address, retval);
	}

	qty = sizeof(references) / sizeof(references[0]);
	for (i=0; i!=qty; ++i) {
		char * retval = gnksa_trim_references_to_len (
			references[i].in, references[i].len);
		const char * pass_msg =
			!strcmp(references[i].expected_out, retval)
			? "PASS"
			:"FAIL";
		printf ("%s references test #%d (expected len [%d], got [%d]\n", pass_msg, i+1, strlen(references[i].expected_out), strlen(retval));
		g_free (retval);
	}
}
