/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#ifndef __GUI_H__
#define __GUI_H__

#include <gtk/gtkwidget.h>
#include <libgnomeui/gnome-app.h>
#include <libgnomeui/gnome-app-helper.h>

#include "article.h"
#include "group.h"
#include "status-item.h"

extern GtkWidget * groups_vbox;
extern GtkWidget * articlelist_ctree;
extern GtkWidget * text_box;
extern GtkWidget * contents_vbox;

extern GnomeUIInfo help_menu[];

enum {
	GUI_PANED = 1,
	GUI_NOTEBOOK
};

enum {
	GROUPS_PAGE = 0,
	ARTICLELIST_PAGE,
	MESSAGE_PAGE
};

void gui_about (void);

void gui_shutdown (void);

void widget_set_font (GtkWidget* w, const char* font_name);

void dialog_draw_centered (GtkWidget *dialog, GtkWidget *window);

void gui_set_layout(int);

void gui_construct (const gchar *geometry);

void gui_page_set (int page, GtkWidget *focus_item);

void gui_set_title (const Group*);

void gui_popup_draw (GtkWidget * dialog, GtkWidget * parent);

void gui_restore_column_widths (GtkWidget * clist, const gchar * type);

void gui_save_column_widths (GtkWidget * clist, const gchar * type);

void gui_select_all (void);

GtkWidget* server_menu_create (void);


/**
 * Adds the specified StatusItem to the main window's list of StatusItems to
 * track.  This will ensure that Pan's main window will keep the user informed
 * of this StatusItem's progress.
 *
 * @see gui_remove_status_item
 * @param StatusItem the StatusItem to track
 */
void gui_add_status_item (StatusItem*);

/**
 * Removes the specified StatusItem from the main window.
 *
 * @see gui_add_status_item
 * @param StatusItem the StatusItem to remove from the main window.
 */
void gui_remove_status_item (StatusItem*);

void gui_set_queue_size (int size);

extern GnomeUIInfo group_menu [];
extern GnomeUIInfo folder_menu [];

extern GdkColormap * cmap;

#endif /* __GUI_H__ */
