/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#ifndef __NNTP_H__
#define __NNTP_H__

#include <glib.h>

#include "server.h"
#include "group.h"
#include "article.h"
#include "sockets.h"
#include "status-item.h"

extern gboolean online_prompt;

void nntp_disconnect       (PanSocket       * socket);

const gchar* nntp_login    (PanSocket       * socket,
                            Server          * server);

gint nntp_set_group        (StatusItem      * status,
                            PanSocket       * socket,
                            const char      * groupname);

gint nntp_post             (PanSocket       * socket,
                            const char      * message);

gint nntp_send_noop        (PanSocket       * socket);

gint nntp_get_group_info   (StatusItem      * item,
                            PanSocket       * sock,
                            const char      * group_name,
                            gulong          * setme_article_qty,
                            gulong          * setme_low_num,
                            gulong          * setme_high_num,
                            const gboolean  * abort );

gint nntp_article_download (StatusItem      * status,
                            PanSocket       * sock,
                            Group           * group,
                            Article         * artilce,
                            const gboolean  * abort_download,
                            gboolean          verbose);

/**
 * This will try to get headers for articles [low-high] from the specified
 * server/group and store them as Article objects in the setme array.
 *
 * If verbose is true and the operation is successful,  the StatusItem will
 * emit low-high status_item_next_step signals.
 */
gint nntp_articlelist_download (StatusItem       * item,
                                PanSocket        * sock,
                                Group            * group,
                                gulong             low,
                                gulong             high,
                                const gboolean   * abort,
			        const char       * progress_fmt,
                                GPtrArray        * setme);

/**
 * emits articles->len status_item_next_step signals.
 */
gint nntp_download_bodies (StatusItem        * item,
                           Group             * group,
                           PanSocket         * sock,
                           const gboolean    * abort,
                           const GPtrArray   * const articles,
                           gint               * index);

#endif /* __NNTP_H__ */
