/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <gnome.h>

#include <stdlib.h>
#include <string.h>

#include "acache.h"
#include "articlelist.h"
#include "globals.h"
#include "gui.h"
#include "gui-headers.h"
#include "gui-paned.h"
#include "nntp.h"
#include "prefs.h"
#include "queue.h"
#include "task-headers.h"
#include "text.h"
#include "util.h"

#include "xpm/traffic_green.xpm"
#include "xpm/traffic_red.xpm"


#define DISPLAY_PREFS_PREVIEW_FONT_SIZE 12


/* server profile */
typedef struct
{
	GtkWidget * profile_name_entry;
	GtkWidget * server_address;
	GtkWidget * server_port;
	GtkWidget * auth_cbutton;
	GtkWidget * server_username_label;
	GtkWidget * server_username;
	GtkWidget * server_password_label;
	GtkWidget * server_password;
	GtkWidget * msgid_cbutton;
	GtkWidget * max_connections;
	GtkWidget * idle_secs_before_timeout;
	Server * server;
}
server_profile;

typedef struct
{
	const Server * server;
	GtkWidget * max_connections_sb;
	GtkWidget * online_button;
	GtkWidget * reserve_tb;
	gboolean online_value;
}
server_connection;

typedef struct
{
	GnomePropertyBox * box;
	GtkWidget * server_clist;

	/* smtp server*/
	GtkWidget * smtp_address;
	GtkWidget * smtp_port;

	/* cache */
	GtkWidget * cache_kbytes_entry;
	GtkWidget * flush_cache_on_exit_check;

	/* general */
	GtkWidget * online_prompt_first_cbutton;
	GtkWidget * online_without_asking_cbutton;
	GtkWidget * online_never_cbutton;
	GtkWidget * one_click_preview_cbutton;
	GtkWidget * hide_mpart_child_nodes_cbutton;
	GtkWidget * attribution_entry;
	GtkWidget * external_mailer_entry;
	GtkWidget * external_editor_entry;
	GtkWidget * auto_part_selection;

	/* user */
	GtkWidget * user_email;
	GtkWidget * user_reply_to;
	GtkWidget * user_organization;
	GtkWidget * user_fullname;
	GtkWidget * user_signature_file;

	/* wrapping */
	GtkWidget * fill_body_on_show;
	GtkWidget * fill_body_on_reply;
	GtkWidget * fill_body_on_send;
	GtkWidget * wrap_column_sb;

	/* directories */
	GtkWidget * dir_download;
	GtkWidget * dir_temp;
	GtkWidget * dir_data;

	/* display */
	GtkWidget * use_gdk_fontset_load_checkbutton;
	GtkWidget * grouplist_gfp;
	GtkWidget * articlelist_gfp;
	GtkWidget * message_body_gfp;

	GtkWidget * read_color_picker;
	GtkWidget * unread_color_picker;
	GtkWidget * new_color_picker;
	GtkWidget * killfile_color_picker;
	GtkWidget * watched_color_picker;
	GtkWidget * text_fg_color_picker;
	GtkWidget * text_bg_color_picker;
	GtkWidget * text_quoted_color_picker;

	/* article headers */
	GtkWidget * article_headers_clist;

	/* connections */
	GtkWidget  * max_connections;
	GtkWidget  * max_tries;
	GSList * server_connections;

}
PrefsWindow;

typedef struct
{
	Server *old;
	Server *new;
}
prefs_server;


static void edit_profile_changed (GnomeDialog *d);
static void prefs_apply_cb (GtkWidget *box, gint page_num);
static void prefs_cancel_cb (void);
static void new_server_cb (void);
static void edit_server_cb (void);
static void delete_server_cb (void);
static void one_click_preview_cbutton_toggled (GtkWidget*, gpointer);
static void prefs_create_clist (void);
static void edit_profile_dialog_clicked (GnomeDialog*, int button, prefs_server*);
static void edit_profile_dialog (prefs_server *data);
static void server_clist_button_press (GtkWidget*, GdkEventButton*);
static void prefs_servers_changed (void);
static GtkWidget * prefs_general_page ( void );
static GtkWidget * prefs_message_page (void);
static GtkWidget * prefs_fill_page (void);
static GtkWidget * prefs_smtp_page (void);
static GtkWidget * prefs_cache_page (void);
static GtkWidget * prefs_server_page (void);
static GtkWidget * prefs_display_page (void);
static GtkWidget * prefs_general_page (void);
static GtkWidget * prefs_directories_page (void);
static GtkWidget * prefs_connections_page (void);
static GtkWidget * prefs_userinfo_page (void);
static gulong get_header_flags (void);

static PrefsWindow * prefs_window;
static gboolean articlelist_font_changed = FALSE;
static gboolean grouplist_changed = FALSE;
static gboolean message_body_changed = FALSE;
static gboolean servers_changed = FALSE;
static GSList * server_deletes = NULL;
static GtkWidget *sections_clist;
static int nntp_clist_row = -1;
static int nntp_connections_row = -1;
static GSList * sprof_list = NULL; /* holds the profile editor GnomeDialogs */



extern GtkTooltips * ttips;


gint watched_thread_span_days = 30;
gint wrap_column = 74;
gboolean fill_body_on_show = FALSE;
gboolean fill_body_on_reply = FALSE;
gboolean fill_body_on_send = FALSE;
gboolean auto_part_selection = TRUE;
gboolean use_gdk_fontset_load = FALSE;
GoingOnlinePreference going_online_preference = OnlinePrompt;
gchar * temp_dir = NULL;
gchar * data_dir = NULL;
gchar * download_dir = NULL;
gchar * grouplist_font = NULL;
gchar * articlelist_font = NULL;
gchar * message_body_font = NULL;
gchar * attribution_line = NULL;
gchar * external_mailer = NULL;
gchar * external_editor = NULL;
gchar * extra_posting_headers = NULL;


/**
***  UPDATE UTILS
**/

static void
update_entry_and_bool_from_toggle_button (gboolean * setme,
                                          const char * key,
                                          GtkWidget * toggle)
{
	const gboolean b = gtk_toggle_button_get_active (
		GTK_TOGGLE_BUTTON(toggle));

	if (b != *setme) {
		*setme = b;
		gnome_config_set_bool (key, b);
	}
}

static void
set_config_from_editable (const char   * key,
                          GtkWidget    * entry)
{
	gchar * text = gtk_editable_get_chars (GTK_EDITABLE(entry), 0, -1);
	g_strstrip (text);
	if (is_nonempty_string(text))
		gnome_config_set_string (key, text);
	else
		gnome_config_clean_key (key);
	g_free (text);
}

static void
replace_string_and_config_from_editable (gchar       ** setme,
                                         const char   * key,
                                         GtkWidget    * e)
{
	gchar * s = gtk_editable_get_chars (GTK_EDITABLE(e), 0, -1);
	g_strstrip (s);
	if (!pan_strcmp(*setme, s))
		g_free (s);
	else {
		gnome_config_set_string (key, s);
		replace_gstr (setme, s);
	}
}

static void
set_color (GdkColor      * color,
           const gchar   * key,
           guint           r,
           guint           g,
           guint           b)
{
	GString * s = g_string_new (NULL);

	g_string_sprintf (s, "/Pan/Display/%s_r=%u", key, r);
	color->red = gnome_config_get_int (s->str);

	g_string_sprintf (s, "/Pan/Display/%s_g=%u", key, g);
	color->green = gnome_config_get_int (s->str);

	g_string_sprintf (s, "/Pan/Display/%s_b=%u", key, b);
	color->blue = gnome_config_get_int (s->str);

	g_string_free (s, TRUE);
}

static gboolean
handle_color_picker (const gchar    * key,
                     GdkColor       * color,
                     GtkWidget      * w)
{
	gboolean color_changed = FALSE;
	gushort r=0, g=0, b=0, a=0;
	gnome_color_picker_get_i16 (GNOME_COLOR_PICKER(w), &r, &g, &b, &a);
	if ((color->red != r)
		|| (color->green != g)
		|| (color->blue != b))
	{
		GString * s = g_string_new (NULL);

		color_changed = TRUE;

		color->red = r;
		color->green = g;
		color->blue = b;

		g_string_sprintf (s, "/Pan/Display/%s_r", key);
		gnome_config_set_int (s->str, r);
		g_string_sprintf (s, "/Pan/Display/%s_g", key);
		gnome_config_set_int (s->str, g);
		g_string_sprintf (s, "/Pan/Display/%s_b", key);
		gnome_config_set_int (s->str, b);
		g_string_sprintf (s, "/Pan/Display/%s_a", key);
		gnome_config_set_int (s->str, a);

		g_string_free (s, TRUE);
	}

	return color_changed;
}

/**
***  UPDATE EVERYTHING
**/



/* Ok or Apply pressed in the Preferences dialog, save all information */
static void
prefs_apply_cb (GtkWidget   * box,
                gint          page_num)
{
	gchar * p;
	int cache_old;
	int cache_new;
	gboolean tmp_bool = FALSE;
	gboolean articlelist_changed = FALSE;
	gboolean server_connections_changed = FALSE;
	gboolean color_changed;

	/* -1 means all pages, ignore anything else */
	if (page_num != -1)
		return;

	/* Commit all of the data to the config file */

	/* No entry for these yet
	 * gnome_config_set_int("/Pan/Articles/Expiration", 30);
	 */

	set_config_from_editable (
		"/Pan/User/Full_Name",
		prefs_window->user_fullname);

	set_config_from_editable (
		"/Pan/User/Email",
		prefs_window->user_email);

	set_config_from_editable (
		"/Pan/User/Reply_To",
		prefs_window->user_reply_to);

	set_config_from_editable (
		"/Pan/User/Organization",
		prefs_window->user_organization);

	set_config_from_editable (
		"/Pan/User/Signature_File",
		gnome_file_entry_gtk_entry(
			GNOME_FILE_ENTRY(prefs_window->user_signature_file)));

	set_config_from_editable (
		"/Pan/Paths/download_dir",
		gnome_file_entry_gtk_entry(
			GNOME_FILE_ENTRY(prefs_window->dir_download)));

	set_config_from_editable (
		"/Pan/Paths/temp_dir",
		gnome_file_entry_gtk_entry(
			GNOME_FILE_ENTRY(prefs_window->dir_temp)));

	set_config_from_editable (
		"/Pan/Paths/data_dir",
		gnome_file_entry_gtk_entry(
			GNOME_FILE_ENTRY(prefs_window->dir_data)));

	set_config_from_editable (
		"/Pan/Mail/smtp_address",
		prefs_window->smtp_address);

	set_config_from_editable (
		"/Pan/Mail/smtp_port",
		prefs_window->smtp_port);


	handle_color_picker (
		"unread_color",
		&unread_color, prefs_window->unread_color_picker);

	handle_color_picker (
		"read_color",
		&read_color, prefs_window->read_color_picker);

	handle_color_picker (
		"new_color",
		&new_color, prefs_window->new_color_picker);

	handle_color_picker (
		"killfile_color",
		&killfile_color, prefs_window->killfile_color_picker);

	handle_color_picker (
		"watched_color",
		&watched_color, prefs_window->watched_color_picker);

	/* update the text window colors */
	color_changed = FALSE;
	color_changed |= handle_color_picker (
		"text_fg_color",
		&text_fg_color, prefs_window->text_fg_color_picker);

	color_changed |= handle_color_picker (
		"text_bg_color",
		&text_bg_color, prefs_window->text_bg_color_picker);

	color_changed |= handle_color_picker (
		"text_quoted_color",
		&text_quoted_color, prefs_window->text_quoted_color_picker);

	if (color_changed) {
		/* must use a new style* in order to update bg color */
		GtkStyle *style=gtk_style_copy(gtk_widget_get_style(Pan.text));
		style->text[0] = text_fg_color;
		style->base[0] = text_bg_color;
		gtk_widget_set_style (Pan.text, style);
		text_refresh ();
	}


	update_entry_and_bool_from_toggle_button (
		&use_gdk_fontset_load,
		"/Pan/General/use_gdk_fontset_load",
		prefs_window->use_gdk_fontset_load_checkbutton);
	
	/* Handle font pickers */
	if (articlelist_font_changed)
	{
		const char * color_err = _("couldn't allocate color");
		const char * cp = gnome_font_picker_get_font_name(
			GNOME_FONT_PICKER(prefs_window->articlelist_gfp));
		gnome_config_set_string ("/Pan/Display/articlelist_font", cp);
		replace_gstr (&articlelist_font, g_strdup(cp));
		articlelist_font_changed = FALSE;

		/* clean out the old styles */
		gtk_style_unref (new_style);
		new_style = NULL;
		gtk_style_unref (read_style);
		read_style = NULL;
		gtk_style_unref (unread_style);
		unread_style = NULL;
		gtk_style_unref (killfile_style);
		killfile_style = NULL;

		/* Make the color changes take effect */
		cmap = gdk_colormap_get_system ();
		if (!gdk_color_alloc (cmap, &unread_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &read_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &new_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &killfile_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &watched_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &text_bg_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &text_fg_color))
			g_warning (color_err);
		if (!gdk_color_alloc (cmap, &text_quoted_color))
			g_warning (color_err);

		articlelist_changed = TRUE;
	}

	if (grouplist_changed)
	{
		const char * cp = gnome_font_picker_get_font_name (
			GNOME_FONT_PICKER(prefs_window->grouplist_gfp));
		gnome_config_set_string (
			"/Pan/Display/grouplist_font", cp);
		replace_gstr (&grouplist_font, g_strdup(cp));
		grouplist_changed = FALSE;

		/* make the new settings take effect */
		widget_set_font (GTK_WIDGET(Pan.group_clist), grouplist_font);
	}

	if (message_body_changed)
	{
		p = gnome_font_picker_get_font_name (GNOME_FONT_PICKER(
				prefs_window->message_body_gfp));
		gnome_config_set_string (
			"/Pan/Display/message_body_font", p);
		replace_gstr (&message_body_font, g_strdup(p));
		message_body_changed = FALSE;

		/* make the new settings take effect */
		widget_set_font (
			GTK_WIDGET(Pan.text), message_body_font);
		text_refresh ();
	}

	/* cache */
	cache_old = gnome_config_get_int_with_default(
		"/Pan/Cache/MaxSize=1242880", NULL);
	cache_new = atoi(gtk_entry_get_text(
			GTK_ENTRY(prefs_window->cache_kbytes_entry)));
	if (cache_old != cache_new) {
		gnome_config_set_int ("/Pan/Cache/MaxSize", cache_new);
		acache_expire ();
	}
	gnome_config_set_bool ("/Pan/Cache/FlushOnExit",
			gtk_toggle_button_get_active (
				GTK_TOGGLE_BUTTON( prefs_window->flush_cache_on_exit_check)));

	/* article headers */
	if (1) {
		gulong flags = get_header_flags ( );
		if (header_flags != flags) {
			gnome_config_set_int ("/Pan/State/Headers", flags);
			header_flags = flags;
			text_refresh ();
		}
	}
	
	if (servers_changed) {
		/* Handle the server mess */
		prefs_servers_changed ();
	}

	/**
	***  Filling
	**/

	update_entry_and_bool_from_toggle_button (
		&auto_part_selection,
		"/Pan/General/auto_part_selection",
		prefs_window->auto_part_selection);

	update_entry_and_bool_from_toggle_button (
		&fill_body_on_show,
		"/Pan/General/fill_body_on_show",
		prefs_window->fill_body_on_show);

	update_entry_and_bool_from_toggle_button (
		&fill_body_on_reply,
		"/Pan/General/fill_body_on_reply",
		prefs_window->fill_body_on_reply);

	update_entry_and_bool_from_toggle_button (
		&fill_body_on_send,
		"/Pan/General/fill_body_on_send",
		prefs_window->fill_body_on_send);

	tmp_bool = hide_mpart_child_nodes;
	update_entry_and_bool_from_toggle_button (
		&hide_mpart_child_nodes,
		"/Pan/General/hide_mpart_child_nodes",
		prefs_window->hide_mpart_child_nodes_cbutton);
	if (tmp_bool != hide_mpart_child_nodes)
		articlelist_changed = TRUE;


	if (1) {
		GtkSpinButton* w = GTK_SPIN_BUTTON(prefs_window->wrap_column_sb);
		int i = gtk_spin_button_get_value_as_int (w);
		if (i != wrap_column) {
			wrap_column = i;
			gnome_config_set_int ("/Pan/General/Wrap_Column", i);
			text_refresh ();

		}
	}

	/*FIXME: there needs to be a way to edit this in the UI */
	gnome_config_set_string ("/Pan/Posting/Extra_Headers", extra_posting_headers);

	replace_string_and_config_from_editable (
		&attribution_line,
		"/Pan/User/attribution_line",
		prefs_window->attribution_entry);

	replace_string_and_config_from_editable (
		&external_mailer,
		"/Pan/User/external_mailer",
		prefs_window->external_mailer_entry);

	replace_string_and_config_from_editable (
		&external_editor,
		"/Pan/User/external_editor",
		prefs_window->external_editor_entry);


	/* server connections ui */
	if (1)
	{
		GSList * l;
		for (l=prefs_window->server_connections; l!=NULL; l=l->next)
		{
			server_connection * sc = (server_connection*) l->data;
			Server * server = SERVER(sc->server);
			int i;

			/* max connections for this server */
			i = gtk_spin_button_get_value_as_int (
				GTK_SPIN_BUTTON(sc->max_connections_sb));
			if (i != server->max_connections)
			{
				server->max_connections = i;
				server_connections_changed = TRUE;
			}

			/* online radiobox */
			i = sc->online_value;
			if ((i!=0) != (server->online!=0))
			{
				server_set_online (server, i);
				server_connections_changed = TRUE;
			}

			/* reserve nonleech */
			i = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(sc->reserve_tb));
			if ((i!=0) != (server->reserve_nonleech!=0))
			{
				server->reserve_nonleech = i;
				server_connections_changed = TRUE;
			}
		}
	}

	/* how to go online */
	if (1)
	{
		GoingOnlinePreference pref;
		GtkToggleButton * never = GTK_TOGGLE_BUTTON (prefs_window->online_never_cbutton);
		GtkToggleButton * always = GTK_TOGGLE_BUTTON (prefs_window->online_without_asking_cbutton);

		if (gtk_toggle_button_get_active (never))
			pref = OnlineNo;
		else if (gtk_toggle_button_get_active (always))
			pref = OnlineYes;
		else
			pref = OnlinePrompt;

		if (pref != going_online_preference)
		{
			going_online_preference = pref;
			gnome_config_set_int ("/Pan/General/Going_Online", (int)pref);
		}
	}

	/* maximum number of connections */
	if (1) {
		GtkSpinButton * sb = GTK_SPIN_BUTTON (prefs_window->max_connections);
		const int i = gtk_spin_button_get_value_as_int (sb);
		if (i != queue_get_max_sockets()) {
			queue_set_max_sockets (i);
			server_connections_changed = TRUE;
			gnome_config_set_int ("/Pan/General/Max_Connections_Total", i);
		}
	}

	/* maximum number of tries */
	if (1) {
		GtkSpinButton * sb = GTK_SPIN_BUTTON (prefs_window->max_tries);
		const int i = gtk_spin_button_get_value_as_int (sb);
		if (i != queue_get_max_tries()) {
			queue_set_max_tries (i);
			server_connections_changed = TRUE;
			gnome_config_set_int ("/Pan/General/Max_Task_Tries_Total", i);
		}
	}

	g_message ("articlelist_changed: %d", articlelist_changed);
	if (articlelist_changed)
		articlelist_reload ();

	if (server_connections_changed)
	{
		queue_wakeup ();
		server_menu_update ();
	}

	gnome_config_sync();

	prefs_init ();
}

static void
prefs_servers_changed (void)
{
	gint row = 0;
	prefs_server * p_server;
	GString * path = g_string_new (NULL);

	while ((p_server = gtk_clist_get_row_data (GTK_CLIST(prefs_window->server_clist), row)))
	{
		if (p_server->new)
		{
			if (p_server->old)
			{
				/* Modified */
				Server * new = p_server->new;
				Server * old = p_server->old;

				replace_gstr (&old->name, g_strdup(new->name));
				replace_gstr (&old->address, g_strdup(new->address));
				replace_gstr (&old->username, g_strdup(new->username));
				replace_gstr (&old->password, g_strdup(new->password));
				old->port = new->port;
				old->need_auth = new->need_auth;
				old->max_connections = new->max_connections;
				old->idle_secs_before_timeout = new->idle_secs_before_timeout;
				old->gen_msgid = new->gen_msgid;
			}
			else
			{
				/* New addition */
				Server * new = p_server->new;
				g_string_sprintf (path, "/%s/%s/", data_dir, new->name);
				directory_check (path->str);
				Pan.serverlist = g_slist_append (Pan.serverlist, new);
				p_server->new = NULL;
			}
		}

		++row;
	}

	if (server_deletes != NULL)
	{
		GSList * l;

		/* remove from server list */
		for (l=server_deletes; l!=NULL; l=l->next)
		{
			prefs_server * p_server = l->data;
			Server * server= p_server->old;
			GSList * m;
			GSList ** pm;

			/* remove it from the pan server list */
			Pan.serverlist = g_slist_remove (Pan.serverlist, server);

			/* remove from the prefs connections tab */
			pm = &prefs_window->server_connections;
			for (m=*pm; m!=NULL; m=m->next)
			{
				server_connection * sc = (server_connection*) m->data;
				if (sc->server == server)
				{
					*pm = g_slist_remove (*pm, sc);
					g_free (sc);
					break;
				}
			}

			/* destroy the server */
			pan_object_unref (PAN_OBJECT(server));
		}

		/* cleanup */
		g_slist_foreach (server_deletes, (GFunc)g_free, NULL);
		g_slist_free (server_deletes);
		server_deletes = NULL;
	}

	server_menu_update ();
	g_string_free (path, TRUE);
}

static void
ensure_trailing_slash (gchar ** pch)
{
	gchar * m = strrchr (*pch, '/');
	if (!m || m[1]!='\0')
		replace_gstr (pch, g_strdup_printf ("%s/", *pch));
}


/**
 * prefs_init:
 *
 * Called from pan_init() before gtk_main(), this is where all the
 * user preferences get loaded in.
 **/
void
prefs_init (void)
{
	static gchar * static_tmp_dir = NULL;
	gchar *pch = NULL;

	/* get temp directory */
	replace_gstr (&temp_dir, gnome_config_get_string ("/Pan/Paths/temp_dir"));
	if (is_nonempty_string (temp_dir))
		ensure_trailing_slash (&temp_dir);
	else
		temp_dir = gnome_util_prepend_user_home (".pan/tmp/");
	directory_check (temp_dir);

	/* get data directory */
	replace_gstr (&data_dir, gnome_config_get_string ("/Pan/Paths/data_dir"));
	if (is_nonempty_string (data_dir))
		ensure_trailing_slash (&data_dir);
	else
		data_dir = gnome_util_prepend_user_home (".pan/data/");
	directory_check (data_dir);

	/* get download directory */
	replace_gstr (&download_dir, gnome_config_get_string ("/Pan/Paths/download_dir"));
	if (is_nonempty_string (download_dir))
		ensure_trailing_slash (&download_dir);
	else
		download_dir = gnome_util_prepend_user_home ("/News/Pan/");
	directory_check (download_dir);


	/* (Set TMPDIR env var, so that uulib uses the right tmpdir) */
	pch = g_strdup_printf ("TMPDIR=%s", temp_dir);
        putenv (pch);
	replace_gstr (&static_tmp_dir, pch); /* purify says not to free pch */
	pch = NULL;

	/* general settings */

	fill_body_on_show = gnome_config_get_bool (
		"/Pan/General/fill_body_on_show=FALSE");
	fill_body_on_reply = gnome_config_get_bool (
		"/Pan/General/fill_body_on_reply=TRUE");
	fill_body_on_send = gnome_config_get_bool (
		"/Pan/General/fill_body_on_send=FALSE");
	wrap_column = gnome_config_get_int (
		"/Pan/General/Wrap_Column=74");
	online_prompt = gnome_config_get_bool (
		"/Pan/General/online_prompt");
	one_click_preview = gnome_config_get_bool (
		"/Pan/General/one_click_preview=FALSE");
	use_gdk_fontset_load = gnome_config_get_bool (
		"/Pan/General/use_gdk_fontset_load=FALSE");
	auto_part_selection = gnome_config_get_bool (
		"/Pan/General/auto_part_selection=TRUE");
	hide_mpart_child_nodes = gnome_config_get_bool (
		"/Pan/General/hide_mpart_child_nodes=TRUE");
	going_online_preference = gnome_config_get_int (
		"/Pan/General/Going_Online=0");
	header_flags = (gulong) gnome_config_get_int (
		"/Pan/State/Headers=243");

	/* user settings */

	replace_gstr (&extra_posting_headers,
		gnome_config_get_string ("/Pan/Posting/Extra_Headers"));
	if (extra_posting_headers == NULL)
		extra_posting_headers = g_strdup ("X-No-Productlinks: Yes");

	replace_gstr (&attribution_line, 
		gnome_config_get_string ("/Pan/User/attribution_line"));
	if (!attribution_line)
		attribution_line = g_strdup (_("In article %i, %a wrote:")),

	replace_gstr (&external_mailer,
		gnome_config_get_string ("/Pan/User/external_mailer"));
	if (!external_mailer)
		external_mailer = g_strdup ("xterm -e mutt -s '%s' -i '%t' '%r'");

	replace_gstr (&external_editor,
		gnome_config_get_string ("/Pan/User/external_editor"));
	if (!external_editor)
		external_editor = g_strdup ("xterm -e vi '%t'");

	/* display settings */

	replace_gstr (&grouplist_font, 
		gnome_config_get_string ("/Pan/Display/grouplist_font"));

	replace_gstr (&articlelist_font,
		gnome_config_get_string ("/Pan/Display/articlelist_font"));

	replace_gstr (&message_body_font,
		gnome_config_get_string ("/Pan/Display/message_body_font"));

	set_color (&unread_color,     "unread_color",         0,     0, 47165);
	set_color (&read_color,       "read_color",        8995,  8995,  8995);
	set_color (&new_color,        "new_color",        65535,  8995,  8995);
	set_color (&killfile_color,   "killfile_color",   35723, 17733,  4883);
	set_color (&watched_color,    "watched_color",        0, 52428,     0);
	set_color (&text_fg_color,    "text_fg_color",        0,     0,     0);
	set_color (&text_bg_color,    "text_bg_color",    65335, 65335, 65335);
	set_color (&text_quoted_color,"text_quoted_color",    0,     0, 65335);
}

static void
prefs_select_row_cb (GtkCList *clist,
		     gint row,
		     gint column,
		     GdkEventButton *event,
		     gpointer user_data)
{
	gtk_notebook_set_page (GTK_NOTEBOOK (prefs_window->box->notebook), row);
}

static int
prefs_add_section_page (gchar     *name,
			GtkWidget *widget)
{
	gchar * text[2];

	text[0] = name;
	text[1] = NULL;

	gtk_notebook_append_page (GTK_NOTEBOOK(prefs_window->box->notebook),
	                          widget,
				  gtk_label_new (name));
	return gtk_clist_append (GTK_CLIST (sections_clist), text);

}

static void
prefs_create_clist (void)
{
	GtkCList * list = GTK_CLIST(prefs_window->server_clist);
	GSList * l;

	/* build the clist */
	gtk_clist_freeze (list);
	gtk_clist_clear (list);
	for (l=Pan.serverlist; l!=NULL; l=l->next)
	{
		Server * server = SERVER(l->data);
		prefs_server * p_server;
		gchar * text[2];
		int row;

		/* skip the internal folder... */
		if (!pan_strcmp (server->name, INTERNAL_SERVER_NAME))
			continue;

		p_server = g_new0 (prefs_server, 1);
		p_server->old = server;
		text[0] = server->name ? server->name : _("Not Named");
		text[1] = server->address ? server->address : _("No Address");
		row = gtk_clist_prepend (list, text);
		gtk_clist_set_row_data (list, row, p_server);
	}
	gtk_clist_thaw (list);
}


static void
edit_profile_dialog_clicked (GnomeDialog     * dialog,
                             int               button,
                             prefs_server    * p_server)
{
	server_profile *sprof = gtk_object_get_data(GTK_OBJECT(dialog),"sprof");

	g_assert (sprof);

	if (button == 0)
	{
		gint row;
		Server *server;

		/* OK button */
		server = server_new ();
		server->name = gtk_editable_get_chars (GTK_EDITABLE (sprof->profile_name_entry), 0, -1);
		server->address = gtk_editable_get_chars (GTK_EDITABLE (sprof->server_address), 0, -1);
		server->port = atoi((gtk_entry_get_text (GTK_ENTRY(sprof->server_port))));
		server->need_auth = (!GTK_TOGGLE_BUTTON (sprof->auth_cbutton)->active) ? 0 : 1;
		server->gen_msgid = (!GTK_TOGGLE_BUTTON (sprof->msgid_cbutton)->active) ? 0 : 1;
		server->username = gtk_editable_get_chars (GTK_EDITABLE (sprof->server_username), 0, -1);
		server->password = gtk_editable_get_chars (GTK_EDITABLE (sprof->server_password), 0, -1);
		server->max_connections = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(sprof->max_connections));
		server->idle_secs_before_timeout = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (sprof->idle_secs_before_timeout));
		server->groups = g_ptr_array_new ();

		if (p_server == NULL)
		{
			gchar *text[2];
			GtkCList * list = GTK_CLIST(prefs_window->server_clist);

			p_server = g_new0 (prefs_server, 1);
			p_server->new = server;
			text[0] = server->name ? server->name : _("Not Named");
			text[1] = server->address ? server->address : _("No Address");
			row = gtk_clist_append (list, text);
			gtk_clist_set_row_data (list, row, p_server);
		}
		else
		{
			GtkCList * list = GTK_CLIST(prefs_window->server_clist);
			if (p_server->new)
				pan_object_unref (PAN_OBJECT(p_server->new));
			p_server->new = server;
			row = gtk_clist_find_row_from_data (list, p_server);
			gtk_clist_set_text (list, row, 1, server->address);
			gtk_clist_set_text (list, row, 0, server->name);
		}
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	sprof_list = g_slist_remove (sprof_list, dialog);
}


static void
edit_profile_changed (GnomeDialog *d)
{
	servers_changed = TRUE;
	gnome_dialog_set_sensitive (d, 0, TRUE);
	gnome_property_box_changed (prefs_window->box);
}


static void
auth_cbutton_set (GtkWidget* widget, server_profile *sprof)
{
	servers_changed = TRUE;
	if (!GTK_TOGGLE_BUTTON (widget)->active) {
		gtk_widget_set_sensitive (sprof->server_username_label,0);
		gtk_widget_set_sensitive (sprof->server_username,0);
		gtk_widget_set_sensitive (sprof->server_password_label,0);
		gtk_widget_set_sensitive (sprof->server_password,0);
	} else {
		/* disable everything */
		gtk_widget_set_sensitive (sprof->server_username_label,1);
		gtk_widget_set_sensitive (sprof->server_username,1);
		gtk_widget_set_sensitive (sprof->server_password_label,1);
		gtk_widget_set_sensitive (sprof->server_password,1);
	}
}


/*---[ edit_profile_dialog ]------------------------------------------
 * The dialog for editing a server profile.
 *--------------------------------------------------------------------*/
static void
edit_profile_dialog (prefs_server *p_server)
{
	GtkWidget *dialog;
	GtkWidget *table;
	GtkWidget *label;
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkAdjustment *adj;
	Server *server;
	server_profile *sprof;

	if (p_server)
		if (p_server->new)
			server = p_server->new;
		else
			server = p_server->old;
	else 
		server = NULL;

	sprof = g_new0 (server_profile, 1);
	
	dialog = gnome_dialog_new (_("New/Edit Server"),
				   GNOME_STOCK_BUTTON_OK,
				   GNOME_STOCK_BUTTON_CANCEL, NULL);
	
	gtk_object_set_data_full (GTK_OBJECT(dialog), "sprof", sprof, g_free);

	gnome_dialog_set_sensitive (GNOME_DIALOG(dialog), 0, FALSE);

	gtk_signal_connect (GTK_OBJECT(dialog), "clicked",
			    GTK_SIGNAL_FUNC (edit_profile_dialog_clicked),
			    p_server);

	table = gtk_table_new (4, 3, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);
	gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), table);

/* FRAME: Server Information */
	frame = gtk_frame_new (_("Server Information"));
	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);
	gtk_container_add (GTK_CONTAINER (frame), vbox);

	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  /* Profile Name */
	label = gtk_label_new (_("Profile Name"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);

	sprof->profile_name_entry = gtk_entry_new ();
	if (server && server->name)
		gtk_entry_set_text (GTK_ENTRY(sprof->profile_name_entry),
				    server->name);
	gtk_signal_connect_object (GTK_OBJECT(sprof->profile_name_entry),
				   "changed",
				   GTK_SIGNAL_FUNC(edit_profile_changed),
				   GTK_OBJECT(dialog));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->profile_name_entry, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  /* Server Address */
	label = gtk_label_new (_("Server Address"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);

	sprof->server_address = gtk_entry_new ();
	if (server && server->address)
		gtk_entry_set_text (GTK_ENTRY(sprof->server_address),
				    server->address);
	gtk_signal_connect_object (GTK_OBJECT(sprof->server_address),
				   "changed",
				   GTK_SIGNAL_FUNC(edit_profile_changed),
				   GTK_OBJECT(dialog));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_address, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  /* Server Port */
	label = gtk_label_new (_("Server Port"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);

	sprof->server_port = gtk_entry_new ();
	if (server && (server->port != 0))
	{
		char port_str [32];
		sprintf (port_str, "%d", server->port);
		gtk_entry_set_text (GTK_ENTRY(sprof->server_port), port_str);
	}
	else
		gtk_entry_set_text (GTK_ENTRY(sprof->server_port), "119");

	gtk_signal_connect_object (GTK_OBJECT(sprof->server_port),
				   "changed",
				   GTK_SIGNAL_FUNC(edit_profile_changed),
				   GTK_OBJECT(dialog));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_port, FALSE, FALSE, 0);

	gtk_table_attach (GTK_TABLE(table), frame, 
			  0, 2, 0, 3,
			  GTK_EXPAND | GTK_FILL,
			  GTK_EXPAND | GTK_FILL,
			  0, 0);

/* FRAME: Authorization */
	frame = gtk_frame_new (_("Authorization"));
	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);
	gtk_container_add (GTK_CONTAINER (frame), vbox);
	
  /* needs autorization information checkbox */

	sprof->auth_cbutton = gtk_check_button_new_with_label (_("My server requires my username and password"));
	if (server && server->need_auth)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sprof->auth_cbutton), TRUE);
	else {
	}

	gtk_signal_connect (GTK_OBJECT (sprof->auth_cbutton), "toggled",
			    GTK_SIGNAL_FUNC (auth_cbutton_set), sprof);
	gtk_signal_connect_object (GTK_OBJECT (sprof->auth_cbutton), "toggled",
				   GTK_SIGNAL_FUNC (edit_profile_changed),
				   GTK_OBJECT (dialog));
	gtk_box_pack_start (GTK_BOX (vbox), sprof->auth_cbutton, FALSE, FALSE, 0);


	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  /* Server Username */
	sprof->server_username_label = gtk_label_new (_("Username"));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_username_label, TRUE, TRUE, 0);
	gtk_misc_set_alignment (GTK_MISC (sprof->server_username_label), 1.0, 0.5);

	sprof->server_username = gtk_entry_new ();
	if (server && server->username)
		gtk_entry_set_text (GTK_ENTRY(sprof->server_username),
				    server->username);
	if (server && server->need_auth)
		gtk_widget_set_sensitive (sprof->server_username, 1);
	else {
		gtk_widget_set_sensitive (GTK_WIDGET (sprof->server_username_label), 0);
		gtk_widget_set_sensitive (sprof->server_username, 0);
	}

	gtk_signal_connect_object (GTK_OBJECT(sprof->server_username),
				   "changed",
				   GTK_SIGNAL_FUNC(edit_profile_changed),
				   GTK_OBJECT(dialog));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_username, FALSE, FALSE, 0);


  /* Server Password */
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	sprof->server_password_label = gtk_label_new (_("Password"));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_password_label, TRUE, TRUE, 0);
	gtk_misc_set_alignment (GTK_MISC (sprof->server_password_label), 1.0, 0.5);

	sprof->server_password = gtk_entry_new ();
	gtk_entry_set_visibility (GTK_ENTRY(sprof->server_password),
				  FALSE);
	if (server && server->password)
		gtk_entry_set_text (GTK_ENTRY(sprof->server_password),
				    server->password);
	if (server && server->need_auth)
		gtk_widget_set_sensitive (sprof->server_password, 1);
	else {
		gtk_widget_set_sensitive (GTK_WIDGET (sprof->server_password_label), 0);
		gtk_widget_set_sensitive (sprof->server_password, 0);
	}

	gtk_signal_connect_object (GTK_OBJECT(sprof->server_password),
				   "changed",
				   GTK_SIGNAL_FUNC(edit_profile_changed),
				   GTK_OBJECT(dialog));

	gtk_box_pack_start (GTK_BOX (hbox), sprof->server_password, FALSE, FALSE, 0);

	gtk_table_attach (GTK_TABLE(table), frame, 
			  0, 2, 4, 6,
			  GTK_EXPAND | GTK_FILL,
			  GTK_EXPAND | GTK_FILL,
			  0, 0);
	
	/* FRAME: Misc */
        frame = gtk_frame_new (_("Misc"));
        vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
        gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);
        gtk_container_add (GTK_CONTAINER (frame), vbox);

	/* SPIN: Max Connections */
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	label = gtk_label_new (_("Maximum Connections:"));
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
	adj = (GtkAdjustment *) gtk_adjustment_new (4.0, 1.0, 4.0, 1.0, 1.0, 0.0);
	sprof->max_connections = gtk_spin_button_new (adj, 0, 0);
	if (server && server->max_connections)
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (sprof->max_connections),
					   server->max_connections);

	gtk_signal_connect_object (GTK_OBJECT (adj), "value_changed",
				   GTK_SIGNAL_FUNC (edit_profile_changed),
				   GTK_OBJECT (dialog));
	gtk_box_pack_start (GTK_BOX (hbox), sprof->max_connections, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);


	/* SPIN: Idle Time */
        hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	label = gtk_label_new (_("Idle Seconds Before Timeout:"));
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

	adj = (GtkAdjustment *) gtk_adjustment_new (180.0, 1.0, 360.0, 10.0, 10.0, 0.0);
	sprof->idle_secs_before_timeout = gtk_spin_button_new (adj, 0, 0);

	if (server)  {
		gtk_spin_button_set_value (
			GTK_SPIN_BUTTON (sprof->idle_secs_before_timeout),
			server->idle_secs_before_timeout);
        }

	gtk_signal_connect_object (GTK_OBJECT (adj), "value_changed",
				   GTK_SIGNAL_FUNC (edit_profile_changed),
				   GTK_OBJECT (dialog));

	gtk_box_pack_start (GTK_BOX (hbox), sprof->idle_secs_before_timeout, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

        /* TOGGLE: Generate Message-ID */
        sprof->msgid_cbutton = gtk_check_button_new_with_label (_("Generate Message-ID"));
        if (server && server->gen_msgid) {
                gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sprof->msgid_cbutton), TRUE);
	}
        else {
        }
        
        gtk_signal_connect_object (GTK_OBJECT (sprof->msgid_cbutton), "toggled",
                                   GTK_SIGNAL_FUNC (edit_profile_changed),
                                   GTK_OBJECT (dialog));
        
        gtk_box_pack_start (GTK_BOX (vbox), sprof->msgid_cbutton, FALSE, FALSE, 0);

        gtk_table_attach (GTK_TABLE(table), frame,
                          0, 2, 7, 9,
                          GTK_EXPAND | GTK_FILL,
                          GTK_EXPAND | GTK_FILL,
                          0, 0);

	sprof_list = g_slist_prepend (sprof_list, dialog);

	gtk_widget_show_all (dialog);
}


static void
server_clist_button_press (GtkWidget *widget, GdkEventButton *bevent)
{
	int row;
	int column;
	prefs_server *p_server = NULL;

	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
					      "button_press_event");

	if ((bevent->button) == 1 && (bevent->type == GDK_2BUTTON_PRESS))
	{
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
					      "button_press_event");
		gtk_clist_get_selection_info (GTK_CLIST(widget), bevent->x,
					      bevent->y, &row, &column);
		p_server = gtk_clist_get_row_data (GTK_CLIST(widget), row);
		edit_profile_dialog (p_server);
	}
}


static void
new_server_cb (void)
{
	edit_profile_dialog (NULL);
}


static void
edit_server_cb (void)
{
	GList * list = GTK_CLIST(prefs_window->server_clist)->selection;
	if (list != NULL)
	{
		prefs_server * p_server = gtk_clist_get_row_data (
			GTK_CLIST (prefs_window->server_clist),
			GPOINTER_TO_INT (list->data));
		edit_profile_dialog (p_server);
	}
}


static void
delete_server_cb (void)
{
	GtkCList * clist = GTK_CLIST(prefs_window->server_clist);
	GList * l = clist->selection;
	if (l != NULL)
	{
		const int row = GPOINTER_TO_INT(l->data);
		prefs_server * p_server = gtk_clist_get_row_data (clist, row);

		server_deletes = g_slist_prepend (server_deletes, p_server);
		gtk_clist_remove (clist, row);
		gnome_property_box_changed (prefs_window->box);
		servers_changed = TRUE;
	}
}


/* Cancel button pressed callback */
static void
prefs_cancel_cb (void)
{
	gint r = 0;
	prefs_server * p_server;

	while ((p_server = gtk_clist_get_row_data (GTK_CLIST(prefs_window->server_clist), r))) {
		if (p_server->new)
			pan_object_unref (PAN_OBJECT(p_server->new));
		g_free (p_server);
		++r;
	}

	/* cleanup widgets */
	g_slist_foreach (sprof_list, (GFunc)gtk_widget_destroy, NULL);
	g_slist_free (sprof_list);
	sprof_list = NULL;

	/* cleanup main window */
	gtk_widget_destroy (GTK_WIDGET(prefs_window->box));
	g_free (prefs_window);
	prefs_window = NULL;
}

static void
one_click_preview_cbutton_toggled (GtkWidget *widget, gpointer data)
{
	const gboolean b = GTK_TOGGLE_BUTTON(widget)->active;
	gnome_property_box_changed (prefs_window->box);

	gnome_config_set_bool ("/Pan/General/one_click_preview", b);
	one_click_preview = b;
}


/*--------------------------------------------------------------------
 * 'general' properties: toggles and such that relate to general
 * functionality and operation of Pan
 *--------------------------------------------------------------------*/
static GtkWidget*
prefs_general_page (void)
{
	int row = 0;
	GtkWidget *w;
	GtkWidget *table;
	table = gtk_table_new (16, 1, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);


	/* single click previewing toggle */
	prefs_window->one_click_preview_cbutton =
		gtk_check_button_new_with_label (
			_("Use single click message "
			  "previewing (in paned layout)"));
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON (prefs_window->one_click_preview_cbutton),
		one_click_preview);
	gtk_signal_connect (
		GTK_OBJECT (prefs_window->one_click_preview_cbutton),
		"toggled",
		GTK_SIGNAL_FUNC (one_click_preview_cbutton_toggled),
		NULL);

	/* auto part selection */
	w = prefs_window->auto_part_selection =
		gtk_check_button_new_with_label (
			_("Automatically guess a part when duplicate parts "
			  "appear in a multipart post"));
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON(w), auto_part_selection);
	gtk_signal_connect_object (
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));

	/* hide 2..n of multipart nodes */
	w = prefs_window->hide_mpart_child_nodes_cbutton =
		gtk_check_button_new_with_label (
			_("Show only the first part of "
			  "complete multipart binary posts"));
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON(w), hide_mpart_child_nodes);
	gtk_signal_connect_object (
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));

	row = 0;

	gtk_table_attach (GTK_TABLE(table),
			  prefs_window->one_click_preview_cbutton,
			  0, 1, row, row+1,
			  GTK_FILL|GTK_EXPAND, GTK_FILL,
			  4, 0);

	++row;

	gtk_table_attach (GTK_TABLE(table),
			  prefs_window->auto_part_selection,
			  0, 1, row, row+1,
			  GTK_FILL|GTK_EXPAND, GTK_FILL,
			  4, 0);
	++row;

	gtk_table_attach (GTK_TABLE(table),
			  prefs_window->hide_mpart_child_nodes_cbutton,
			  0, 1, row, row+1,
			  GTK_FILL|GTK_EXPAND, GTK_FILL,
			  4, 0);

	++row;

	w = gtk_hseparator_new();
	gtk_table_attach (
		GTK_TABLE(table), w,
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, GNOME_PAD_BIG);

	++row;

	w = gtk_label_new (
		_("Attribution Line (%i for message-id, %a for author, %d for date)"));
	gtk_misc_set_alignment (
		GTK_MISC(w), 0.0, 0.5);
	gtk_table_attach (
		GTK_TABLE(table), w,
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, 0);

	++row;

	prefs_window->attribution_entry = gtk_entry_new();
	gtk_entry_set_text (GTK_ENTRY(prefs_window->attribution_entry), attribution_line);
        gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->attribution_entry), "changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));
	gtk_table_attach (
		GTK_TABLE(table),
		prefs_window->attribution_entry,
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, 0);

	++row;

	w = gtk_hseparator_new();
	gtk_table_attach (
		GTK_TABLE(table), w,
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, GNOME_PAD_BIG);

	++row;

	gtk_table_attach (GTK_TABLE(table),
			  gtk_label_new(_("External Mail Program (%r for recipient, %s for subject, %t for body filename)")),
			  0, 1, row, row+1,
			  GTK_FILL|GTK_EXPAND, GTK_FILL,
			  0, 0);

	++row;

	prefs_window->external_mailer_entry = gtk_entry_new();
	gtk_entry_set_text (GTK_ENTRY(prefs_window->external_mailer_entry), external_mailer);
        gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->external_mailer_entry), "changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));
	gtk_table_attach (GTK_TABLE(table),
			  prefs_window->external_mailer_entry,
			  0, 1, row, row+1,
			  GTK_FILL|GTK_EXPAND, GTK_FILL,
			  0, 0);

	++row;

	gtk_table_attach (
		GTK_TABLE(table),
		gtk_label_new(_("External Editor (%t for temporary filename)")),
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, 0);

	++row;

	prefs_window->external_editor_entry = gtk_entry_new();
	gtk_entry_set_text (
		GTK_ENTRY(prefs_window->external_editor_entry), external_editor);
        gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->external_editor_entry), "changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));
	gtk_table_attach (
		GTK_TABLE(table),
		prefs_window->external_editor_entry,
		0, 1, row, row+1,
		GTK_FILL|GTK_EXPAND, GTK_FILL,
		0, 0);

	++row;

	return table;
}

static void
set_grouplist_changed (void)
{
	grouplist_changed = TRUE;
}

static void
set_articlelist_font_changed (void)
{
	articlelist_font_changed = TRUE;
}

static void
set_message_body_changed (void)
{
	message_body_changed = TRUE;
}

static void
reset_wrap_column_cb (void)
{
	GtkSpinButton * w = GTK_SPIN_BUTTON (prefs_window->wrap_column_sb);
	gtk_spin_button_set_value (w, 74);
}

/**
 * Build and return the 'Fill' preferences page;
 * lets users decide when message bodies are filled
 */
static GtkWidget*
prefs_fill_page (void)
{
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *w;
	GtkAdjustment *adj;

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GNOME_PAD_SMALL);
	w = gtk_label_new (_("``Filling'' an article reformats its text and wraps it to fit\nin an 80-column display, as is standard practice on Usenet."));
	gtk_box_pack_start (GTK_BOX(vbox), w, FALSE, FALSE, GNOME_PAD_BIG);

	/* fill body on show */
	w = prefs_window->fill_body_on_show = gtk_check_button_new_with_label (
		_("Fill articles for reading"));
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON(w), fill_body_on_show);
	gtk_signal_connect_object (
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(vbox), w, FALSE, FALSE, 0);

	/* fill body on reply */
	w = prefs_window->fill_body_on_reply = gtk_check_button_new_with_label (
		_("Fill the quoted article when replying/following-up"));
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON(w), fill_body_on_reply);
	gtk_signal_connect_object (
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(vbox), w, FALSE, FALSE, 0);

	/* fill body on send */
	w = prefs_window->fill_body_on_send = gtk_check_button_new_with_label (
		_("Fill your post, follow-up, or reply when it's sent.\n"
		  "This is disabled until fill's been tested more.  For the\n"
		  "time being, use fill button located in the compose window."));
	gtk_misc_set_alignment (
		GTK_MISC(GTK_BIN(w)->child), 0.0, 0.0);
	gtk_toggle_button_set_active (
		GTK_TOGGLE_BUTTON(w), fill_body_on_send);
	gtk_signal_connect_object (
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_widget_set_sensitive (w, FALSE);
	gtk_box_pack_start (GTK_BOX(vbox), w, FALSE, FALSE, 0);

	/* where to fill to */
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(hbox), gtk_label_new (_("Wrap Articles at Column")), FALSE, FALSE, 0);
	adj = (GtkAdjustment *) gtk_adjustment_new (256.0, 1.0, 256.0, 1.0, 1.0, 0.0);
	w = prefs_window->wrap_column_sb = gtk_spin_button_new (adj, 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON(w), wrap_column);
	gtk_signal_connect_object (
		GTK_OBJECT (adj), "value_changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);
	w = gtk_button_new_with_label (_("Reset"));
	gtk_signal_connect (GTK_OBJECT(w), "clicked", reset_wrap_column_cb, NULL);
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, GNOME_PAD);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, GNOME_PAD);

	return vbox;
}

/* Build and return the 'Display' preferences page, lets users
 * change fonts and colors
 */
static GtkWidget *
prefs_display_page (void)
{
	GtkWidget *table;
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *w;
	gchar *p;

	/* initialize the variables */
	prefs_window->grouplist_gfp = gnome_font_picker_new ();
	prefs_window->articlelist_gfp = gnome_font_picker_new ();
	prefs_window->message_body_gfp = gnome_font_picker_new ();

	prefs_window->unread_color_picker = gnome_color_picker_new ();
	prefs_window->read_color_picker = gnome_color_picker_new ();
	prefs_window->new_color_picker = gnome_color_picker_new ();
	prefs_window->killfile_color_picker = gnome_color_picker_new ();
	prefs_window->watched_color_picker = gnome_color_picker_new ();
	prefs_window->text_fg_color_picker = gnome_color_picker_new ();
	prefs_window->text_bg_color_picker = gnome_color_picker_new ();
	prefs_window->text_quoted_color_picker = gnome_color_picker_new ();

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);


/**
 * fonts frame
 **/
	table = gtk_table_new (4, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	frame = gtk_frame_new (_("Fonts"));

   /* grouplist font */
	if ((p = gnome_config_get_string ("/Pan/Display/grouplist_font")) != NULL) {
		gnome_font_picker_set_font_name (GNOME_FONT_PICKER (prefs_window->grouplist_gfp), p);
		g_free (p);
	}

	gnome_font_picker_set_mode (GNOME_FONT_PICKER (prefs_window->grouplist_gfp),
				    GNOME_FONT_PICKER_MODE_FONT_INFO);
	gnome_font_picker_fi_set_use_font_in_label (GNOME_FONT_PICKER (prefs_window->grouplist_gfp),
						    TRUE,
						    DISPLAY_PREFS_PREVIEW_FONT_SIZE);

	gtk_signal_connect (GTK_OBJECT(prefs_window->grouplist_gfp), "font_set",
			    GTK_SIGNAL_FUNC(set_grouplist_changed), NULL);

	gtk_signal_connect_object (GTK_OBJECT (prefs_window->grouplist_gfp), "font_set",
				   GTK_SIGNAL_FUNC (gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));

	w = gtk_label_new (_("Group List"));
	gtk_misc_set_alignment (GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (GTK_TABLE (table), prefs_window->grouplist_gfp, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD_SMALL, 0);

	/* articlelist font */
	if ((p = gnome_config_get_string ("/Pan/Display/articlelist_font")) != NULL) {
		gnome_font_picker_set_font_name (GNOME_FONT_PICKER (prefs_window->articlelist_gfp), p);
		g_free (p);
	}

	gnome_font_picker_set_mode (GNOME_FONT_PICKER (prefs_window->articlelist_gfp),
				    GNOME_FONT_PICKER_MODE_FONT_INFO);
	gnome_font_picker_fi_set_use_font_in_label (GNOME_FONT_PICKER (prefs_window->articlelist_gfp),
						    TRUE,
						    DISPLAY_PREFS_PREVIEW_FONT_SIZE);

	gtk_signal_connect (GTK_OBJECT(prefs_window->articlelist_gfp), "font_set",
			    GTK_SIGNAL_FUNC(set_articlelist_font_changed), NULL);

	gtk_signal_connect_object (GTK_OBJECT(prefs_window->articlelist_gfp), "font_set",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));

	w = gtk_label_new (_("Article List"));
	gtk_misc_set_alignment (GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 1, 2,
			  GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (GTK_TABLE (table), prefs_window->articlelist_gfp, 1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD_SMALL, 0);

	/* message body font */
	if ((p = gnome_config_get_string ("/Pan/Display/message_body_font")) != NULL) {
		gnome_font_picker_set_font_name (GNOME_FONT_PICKER (prefs_window->message_body_gfp), p);
		g_free (p);
	}

	gnome_font_picker_set_mode (GNOME_FONT_PICKER (prefs_window->message_body_gfp),
				    GNOME_FONT_PICKER_MODE_FONT_INFO);
	gnome_font_picker_fi_set_use_font_in_label (GNOME_FONT_PICKER(prefs_window->message_body_gfp),
						    TRUE,
						    DISPLAY_PREFS_PREVIEW_FONT_SIZE);

	gtk_signal_connect (GTK_OBJECT(prefs_window->message_body_gfp), "font_set",
			    GTK_SIGNAL_FUNC(set_message_body_changed), NULL);

	gtk_signal_connect_object (GTK_OBJECT(prefs_window->message_body_gfp), "font_set",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));

	w = gtk_label_new (_("Message Body"));
	gtk_misc_set_alignment (GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 2, 3,
			  GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (GTK_TABLE (table), prefs_window->message_body_gfp, 1, 2, 2, 3,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD_SMALL, 0);

	/* how to load fonts? */
	w = gtk_check_button_new_with_label (_("Use gdk_fontset_load instead of gdk_font_load"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w),
	                              use_gdk_fontset_load);
	gtk_signal_connect_object (GTK_OBJECT(w), "toggled",
	                           GTK_SIGNAL_FUNC(gnome_property_box_changed),
	                           GTK_OBJECT (prefs_window->box));
	gtk_table_attach (GTK_TABLE (table), w, 0, 2, 3, 4,
			  GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	prefs_window->use_gdk_fontset_load_checkbutton = w;
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w,
		_("Toggling this and restarting Pan is a good idea if your fonts aren't being displayed properly."), NULL);

	/* pack the font frame */
	gtk_container_add (GTK_CONTAINER (frame), table);
	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, GNOME_PAD_SMALL);

	/**
	*** colors frame
	**/

	frame = gtk_frame_new (_("Colors"));
	table = gtk_table_new (4, 6, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	/* unread article */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->unread_color_picker),
		unread_color.red,
		unread_color.green,
		unread_color.blue,
		gnome_config_get_int ("/Pan/Display/unread_color_a=65534"));
	w = gtk_label_new (_("Unread Message"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		0, 1, 0, 1,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (
		GTK_TABLE (table), prefs_window->unread_color_picker,
		1, 2, 0, 1,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->unread_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* read article */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->read_color_picker),
		read_color.red,
		read_color.green,
		read_color.blue,
		gnome_config_get_int ("/Pan/Display/read_color_a=65534"));
	w = gtk_label_new (_("Read Message"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE(table), w,
		0, 1, 1, 2,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (
		GTK_TABLE (table), prefs_window->read_color_picker,
		1, 2, 1, 2,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->read_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* new article */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->new_color_picker),
		new_color.red,
		new_color.green,
		new_color.blue,
		gnome_config_get_int ("/Pan/Display/new_color_a=65534"));
	w = gtk_label_new (_("New Message"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE(table), w,
		0, 1, 2, 3,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (
		GTK_TABLE(table), prefs_window->new_color_picker,
		1, 2, 2, 3,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->new_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* killfile article */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->killfile_color_picker),
		killfile_color.red,
		killfile_color.green,
		killfile_color.blue,
		gnome_config_get_int ("/Pan/Display/killfile_color_a=65534"));
	w = gtk_label_new (_("Killfile Message"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		0, 1, 3, 4,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (
		GTK_TABLE (table),
		prefs_window->killfile_color_picker,
		1, 2, 3, 4,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->killfile_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* watched thread */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->watched_color_picker),
		watched_color.red,
		watched_color.green,
		watched_color.blue,
		gnome_config_get_int ("/Pan/Display/watched_color_a=65534"));
	w = gtk_label_new (_("Tracked Thread"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		0, 1, 4, 5,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (
		GTK_TABLE (table),
		prefs_window->watched_color_picker,
		1, 2, 4, 5,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->watched_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));
	gtk_container_add (GTK_CONTAINER (frame), table);

	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);

	/* text background */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->text_bg_color_picker),
		text_bg_color.red,
		text_bg_color.green,
		text_bg_color.blue,
		gnome_config_get_int ("/Pan/Display/text_bg_color_a=65534"));
	w = gtk_label_new (_("Article Background"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		2, 3, 0, 1,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (
		GTK_TABLE (table), prefs_window->text_bg_color_picker,
		3, 4, 0, 1,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->text_bg_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* text foreground */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->text_fg_color_picker),
		text_fg_color.red,
		text_fg_color.green,
		text_fg_color.blue,
		gnome_config_get_int ("/Pan/Display/text_fg_color_a=65534"));
	w = gtk_label_new (_("Article Foreground"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		2, 3, 1, 2,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (
		GTK_TABLE (table), prefs_window->text_fg_color_picker,
		3, 4, 1, 2,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->text_fg_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	/* quoted text */
	gnome_color_picker_set_i16 (
		GNOME_COLOR_PICKER (prefs_window->text_quoted_color_picker),
		text_quoted_color.red,
		text_quoted_color.green,
		text_quoted_color.blue,
		gnome_config_get_int ("/Pan/Display/text_quoted_color_a=65534"));
	w = gtk_label_new (_("Article Quoted Text"));
	gtk_misc_set_alignment (
		GTK_MISC (w), 1.0, 0.5);
	gtk_table_attach (
		GTK_TABLE (table), w,
		2, 3, 2, 3,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_table_attach (
		GTK_TABLE (table), prefs_window->text_quoted_color_picker,
		3, 4, 2, 3,
		GTK_FILL, GTK_FILL, GNOME_PAD_SMALL, 0);
	gtk_signal_connect_object (
		GTK_OBJECT(prefs_window->text_quoted_color_picker), "color_set",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT(prefs_window->box));

	return vbox;
}

/*---[ prefs_connections_page ]---------------------------------------
 * Connections preferences page for setting the data, temp, and download
 * dirs in Pan.
 *--------------------------------------------------------------------*/

static GtkWidget*
inset (GtkWidget* w)
{
	GtkWidget* outer;
	GtkWidget* inner;

	inner = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(inner), GTK_SHADOW_NONE);
	gtk_container_set_border_width (GTK_CONTAINER(inner), GNOME_PAD_SMALL);
	gtk_container_add (GTK_CONTAINER(inner), w);

	outer = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(outer), GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER(outer), inner);
	gtk_container_set_border_width (GTK_CONTAINER(inner), 0);

	return outer;
}

static void
online_clicked_cb (GtkButton *b, gpointer user_data)
{
	char * pch;
	server_connection * sc = (server_connection*) user_data;
	GtkWidget * w;

	sc->online_value = !sc->online_value;
	pch = sc->online_value
		? g_strdup_printf (_("Server \"%s\" is Online"), sc->server->name)
		: g_strdup_printf (_("Server \"%s\" is Offline; no new tasks will be processed"), sc->server->name);

	pan_lock ();
       	w = sc->online_button;
	gtk_container_remove (GTK_CONTAINER(w), GTK_BIN(w)->child);
	gtk_container_add (GTK_CONTAINER(w), gnome_pixmap_new_from_xpm_d (sc->online_value?traffic_green_xpm:traffic_red_xpm));
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, pch, "");
	gtk_widget_show_all (w);
	pan_unlock ();

	gnome_property_box_changed (prefs_window->box);

	g_free (pch);
}

static GtkWidget*
prefs_connections_page (void)
{
	GtkAdjustment *adj;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *radiobox;
	GtkWidget *table;
	GtkWidget *frame;
	GtkWidget *w;
	GSList *l;
	int row;
	const int max_per_server =
		gnome_config_get_int ("/Pan/General/Max_Connections_Per_Server=4");

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);

	/* max connections total */
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	w = gtk_label_new (_("Maximum Total Number of Connections"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);
	adj = (GtkAdjustment *) gtk_adjustment_new (128.0, 1.0, 128.0, 1.0, 1.0, 0.0);
	prefs_window->max_connections = gtk_spin_button_new (adj, 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (prefs_window->max_connections), queue_get_max_sockets());
	gtk_signal_connect_object (
		GTK_OBJECT (adj), "value_changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(hbox), prefs_window->max_connections, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	/* task retries */
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	w = gtk_label_new (_("Maximum Number of Task Retries before Giving Up"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);
	adj = (GtkAdjustment *) gtk_adjustment_new (4.0, 1.0, 999.0, 1.0, 1.0, 0.0);
	prefs_window->max_tries = gtk_spin_button_new (adj, 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (prefs_window->max_tries), queue_get_max_tries());
	gtk_signal_connect_object (
		GTK_OBJECT (adj), "value_changed",
		GTK_SIGNAL_FUNC(gnome_property_box_changed),
		GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(hbox), prefs_window->max_tries, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	/* ask before going online */
	radiobox = gtk_vbox_new (FALSE, 0);
	w = gtk_radio_button_new_with_label (NULL, _("Pan should ask before going online"));
	gtk_box_pack_start (GTK_BOX(radiobox), w, FALSE, FALSE, 0);
	prefs_window->online_prompt_first_cbutton = w;
	gtk_signal_connect_object (GTK_OBJECT(w), "toggled", GTK_SIGNAL_FUNC(gnome_property_box_changed), GTK_OBJECT (prefs_window->box));
	w = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(w), _("Pan should go online as needed without asking"));
	gtk_box_pack_start (GTK_BOX(radiobox), w, FALSE, FALSE, 0);
	prefs_window->online_without_asking_cbutton = w;
	gtk_signal_connect_object (GTK_OBJECT(w), "toggled", GTK_SIGNAL_FUNC(gnome_property_box_changed), GTK_OBJECT (prefs_window->box));
	w = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(w), _("Pan should quietly stay offline until told otherwise"));
	gtk_box_pack_start (GTK_BOX(radiobox), w, FALSE, FALSE, 0);
	prefs_window->online_never_cbutton = w;
	gtk_signal_connect_object (GTK_OBJECT(w), "toggled", GTK_SIGNAL_FUNC(gnome_property_box_changed), GTK_OBJECT (prefs_window->box));
	switch (going_online_preference) {
		case OnlineYes: w = prefs_window->online_without_asking_cbutton; break;
		case OnlineNo: w = prefs_window->online_never_cbutton; break;
		case OnlinePrompt:
		default: w = prefs_window->online_prompt_first_cbutton; break;
	}
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w), TRUE);
	gtk_box_pack_start (GTK_BOX(vbox), radiobox, FALSE, FALSE, 0);


	/* pack the vbox into a frame, which is itself the first element in a new vbox */
	frame = gtk_frame_new (_("Global Settings"));
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GNOME_PAD_SMALL);
	gtk_container_add (GTK_CONTAINER(frame), vbox);
	vbox = gtk_vbox_new (FALSE, GNOME_PAD_BIG);
	gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);

	/* per-server settings */
	table = gtk_table_new (g_slist_length(Pan.serverlist)+1, 3, FALSE);
	w = gtk_label_new (_("News Server"));
	gtk_misc_set_padding (GTK_MISC(w), GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 0, 1,  GTK_FILL, GTK_FILL, 0, 0);
	w = gtk_label_new (_("Maximum\nConnections"));
	gtk_misc_set_padding (GTK_MISC(w), GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (GTK_TABLE (table), w, 1, 2, 0, 1,  GTK_FILL, GTK_FILL, 0, 0);
	w = gtk_label_new (_("Online Status"));
	gtk_misc_set_padding (GTK_MISC(w), GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (GTK_TABLE (table), w, 2, 3, 0, 1,  GTK_FILL, GTK_FILL, 0, 0);
	w = gtk_label_new (_("Reserve one connection\nfor non-decode tasks"));
	gtk_misc_set_padding (GTK_MISC(w), GNOME_PAD_SMALL, GNOME_PAD_SMALL);
	gtk_table_attach (GTK_TABLE (table), w, 3, 4, 0, 1,  GTK_FILL, GTK_FILL, 0, 0);
	for (row=1, l=Pan.serverlist; l!=NULL; l=l->next, ++row)
	{
		const Server* server = SERVER(l->data);
		server_connection * sc;
		char * pch;

		/* skip the internal server */
		if (!pan_strcmp(server->name, INTERNAL_SERVER_NAME))
			continue;

		sc = g_new0 (server_connection, 1);
		sc->server = server;

		/* server name */
		w = gtk_label_new (server->name);
		gtk_table_attach (GTK_TABLE(table), inset(w), 0, 1, row, row+1, GTK_FILL, GTK_FILL, 0, 0);

		/* max connections per server */
		adj = (GtkAdjustment *) gtk_adjustment_new (max_per_server, 1.0, max_per_server, 1.0, 1.0, 0.0);
		w = gtk_spin_button_new (adj, 0, 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON(w), server->max_connections);
		gtk_signal_connect_object (
			GTK_OBJECT (adj), "value_changed",
			GTK_SIGNAL_FUNC(gnome_property_box_changed),
			GTK_OBJECT (prefs_window->box));
		gtk_table_attach (GTK_TABLE(table), inset(w), 1, 2, row, row+1, GTK_FILL, GTK_FILL, 0, 0);
		sc->max_connections_sb = w;

		/* online/online */
		w = sc->online_button = gtk_button_new ();
		gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
		sc->online_value = server->online;
		pch = sc->online_value
			? g_strdup_printf (_("Server \"%s\" is Online"), server->name)
			: g_strdup_printf (_("Server \"%s\" is Offline; no new tasks will be processed"), server->name);
		gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, pch, "");
		g_free (pch);
		gtk_container_add (GTK_CONTAINER(w), gnome_pixmap_new_from_xpm_d (server->online?traffic_green_xpm:traffic_red_xpm));
		gtk_signal_connect (GTK_OBJECT(w), "clicked", online_clicked_cb, sc);
		gtk_table_attach (GTK_TABLE(table), inset(w), 2, 3, row, row+1, GTK_FILL, GTK_FILL, 0, 0);

		/* reserve */
		w = gtk_check_button_new_with_label (_("Reserve"));
		sc->reserve_tb = w;
		gtk_table_attach (GTK_TABLE(table), inset(w), 3, 4, row, row+1, GTK_FILL, GTK_FILL, 0, 0);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w), server->reserve_nonleech);
		gtk_signal_connect_object (
			GTK_OBJECT(w), "toggled",
			GTK_SIGNAL_FUNC(gnome_property_box_changed),
			GTK_OBJECT (prefs_window->box));

		/* remember this server */
		prefs_window->server_connections = g_slist_append (prefs_window->server_connections, sc);
	}

	frame = gtk_frame_new (_("Per-Server Settings"));
	gtk_container_set_border_width (GTK_CONTAINER(table), GNOME_PAD_SMALL);
	gtk_container_add (GTK_CONTAINER(frame), table);
	gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);

	return vbox;
}


/*---[ prefs_directories_page ]---------------------------------------
 * Directory preferences page for setting the data, temp, and download
 * dirs in Pan.
 *--------------------------------------------------------------------*/
static GtkWidget *
prefs_directories_page (void)
{
	GtkWidget *table;
	GtkWidget *label;
	gchar *p;

	table = gtk_table_new (3, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);

	label = gtk_label_new (_("Download Directory:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->dir_download = gnome_file_entry_new ("Prefs", _("Download Directory"));
	if ((p = gnome_config_get_string ("/Pan/Paths/download_dir")) == NULL)
		p = gnome_util_prepend_user_home ("News/Pan/");
	gtk_entry_set_text (GTK_ENTRY(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_download))), p);
	gnome_file_entry_set_directory (GNOME_FILE_ENTRY (prefs_window->dir_download), TRUE);

	g_free (p);
	gtk_signal_connect_object (GTK_OBJECT(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_download))),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));
	gtk_table_attach (GTK_TABLE(table), prefs_window->dir_download, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	label = gtk_label_new (_("Temporary Directory:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 1, 2,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->dir_temp = gnome_file_entry_new ("Prefs", _("Temporary Directory"));
	if ((p = gnome_config_get_string("/Pan/Paths/temp_dir")) == NULL) {
		p = gnome_util_prepend_user_home(".pan/tmp/");
	}
	gtk_entry_set_text (GTK_ENTRY(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_temp))), p);
	gnome_file_entry_set_directory (GNOME_FILE_ENTRY (prefs_window->dir_temp), TRUE);
	g_free (p);
	gtk_signal_connect_object (GTK_OBJECT(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_temp))),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));
	gtk_table_attach (GTK_TABLE(table), prefs_window->dir_temp, 1, 2, 1,
			  2, GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

/* Data Directory: */
	label = gtk_label_new (_("Data Directory:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 2, 3,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->dir_data = gnome_file_entry_new ("Prefs", _("Data Directory"));
	if ((p = gnome_config_get_string("/Pan/Paths/data_dir")) == NULL) {
		p = gnome_util_prepend_user_home(".pan/data/");
	}
	gtk_entry_set_text (GTK_ENTRY(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_data))), p);
	gnome_file_entry_set_directory (GNOME_FILE_ENTRY (prefs_window->dir_data), TRUE);
	g_free (p);
	gtk_signal_connect_object (GTK_OBJECT(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->dir_data))),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));
	gtk_table_attach (GTK_TABLE(table), prefs_window->dir_data, 1, 2, 2,
			  3, GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	return table;
}


/*---[ prefs_userinfo_page ]------------------------------------------
 * page that holds the user information such as full name,
 * e-mail, organization, and signature file (this page is optional).
 *--------------------------------------------------------------------*/
static GtkWidget *
prefs_userinfo_page (void)
{
	GtkWidget *table;
	GtkWidget *label;
	char *p;

	table = gtk_table_new (5, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);

	label = gtk_label_new (_("Full Name:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->user_fullname = gtk_entry_new ();
	if ((p = gnome_config_get_string("/Pan/User/Full_Name"))) {
		gtk_entry_set_text (GTK_ENTRY(prefs_window->user_fullname), p);
		g_free(p);
	}
	gtk_signal_connect_object (GTK_OBJECT(prefs_window->user_fullname),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));
	gtk_widget_show (prefs_window->user_fullname);
	gtk_table_attach (GTK_TABLE(table), prefs_window->user_fullname, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	label = gtk_label_new (_("Email Address:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 1, 2,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->user_email = gtk_entry_new ();
	if ((p = gnome_config_get_string("/Pan/User/Email"))) {
		gtk_entry_set_text (GTK_ENTRY(prefs_window->user_email), p);
		g_free(p);
	}
	gtk_signal_connect_object (GTK_OBJECT(prefs_window->user_email),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));
	gtk_widget_show (prefs_window->user_email);
	gtk_table_attach (GTK_TABLE(table), prefs_window->user_email, 1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	label = gtk_label_new (_("Reply To:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
    gtk_table_attach (GTK_TABLE(table), label, 0, 1, 2, 3,
				GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->user_reply_to = gtk_entry_new ();
    if ((p = gnome_config_get_string("/Pan/User/Reply_To"))) {
		gtk_entry_set_text (GTK_ENTRY(prefs_window->user_reply_to), p);
		g_free(p);
	}
	gtk_signal_connect_object (GTK_OBJECT(prefs_window->user_reply_to),
				"changed",
				GTK_SIGNAL_FUNC(gnome_property_box_changed),
				GTK_OBJECT(prefs_window->box));
	gtk_widget_show (prefs_window->user_reply_to);
	gtk_table_attach (GTK_TABLE(table), prefs_window->user_reply_to, 1,
				2, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);
				
	label = gtk_label_new (_("Organization:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 3, 4,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->user_organization = gtk_entry_new ();
	if ((p = gnome_config_get_string("/Pan/User/Organization"))) {
		gtk_entry_set_text (GTK_ENTRY(prefs_window->user_organization), p);
		g_free(p);
	}
	gtk_signal_connect_object (GTK_OBJECT(prefs_window->user_organization),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));
	gtk_widget_show (prefs_window->user_organization);
	gtk_table_attach (GTK_TABLE(table), prefs_window->user_organization, 1,
			  2, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	label = gtk_label_new (_("Signature File:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE(table), label, 0, 1, 4, 5,
			  GTK_FILL, GTK_FILL, 0, 0);

	prefs_window->user_signature_file = gnome_file_entry_new ("Prefs", _("Signature File:"));
	if ((p = gnome_config_get_string("/Pan/User/Signature_File"))) {
		gtk_entry_set_text (GTK_ENTRY(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->user_signature_file))), p);
		g_free(p);
	}
	gtk_signal_connect_object (GTK_OBJECT(gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (prefs_window->user_signature_file))),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));
	gtk_widget_show (GTK_WIDGET (prefs_window->user_signature_file));
	gtk_table_attach (GTK_TABLE(table), GTK_WIDGET (prefs_window->user_signature_file), 1, 2, 4, 5,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);
	return table;
}

static void
header_selection_changed (GtkCList* clist,
			  gint row,
			  gint column,
			  GdkEventButton *event,
			  gpointer user_data)
{
	gnome_property_box_changed (prefs_window->box);
}

static gulong
get_header_flags (void)
{
	GtkCList *clist = GTK_CLIST(prefs_window->article_headers_clist);
	GList* sel = clist->selection;
	gulong flags = 0;

	while ( sel != NULL ) {
		int row_index = GPOINTER_TO_INT ( sel->data );
		gulong* pulong = gtk_clist_get_row_data ( clist, row_index );
		flags |= *pulong;
		sel = sel->next;
	}

	return flags;
}

typedef struct
{
	gchar* text;
	gulong flag;
} MessageListThrowawayStruct;

static GtkWidget *
prefs_message_page (void)
{
	int i;
	GtkWidget * label;
	GtkWidget * vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	GtkCList* list = GTK_CLIST (gtk_clist_new(1));
	gulong flags;
	MessageListThrowawayStruct rows [] = {
		{NULL, UI_HEADER_AUTHOR},
		{NULL, UI_HEADER_DATE},
		{NULL, UI_HEADER_FOLLOWUP_TO},
		{NULL, UI_HEADER_MESSAGE_ID},
		{NULL, UI_HEADER_NEWSGROUPS},
		{NULL, UI_HEADER_REFERENCES},
		{NULL, UI_HEADER_REPLY_TO},
		{NULL, UI_HEADER_SUBJECT}
	};
	const int row_qty = sizeof(rows) / sizeof(rows[0]);
       	flags = (gulong) gnome_config_get_int_with_default (
		"/Pan/State/Headers=243", NULL);
	rows[0].text = _("Author");
	rows[1].text = _("Date");
	rows[2].text = _("Followup-To");
	rows[3].text = _("Message-Id");
	rows[4].text = _("Newsgroups (if more than one group)");
	rows[5].text = _("References (if article is a reply)");
	rows[6].text = _("Reply-To (if different from Author)");
	rows[7].text = _("Subject");

	gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);

	prefs_window->article_headers_clist = GTK_WIDGET(list);
	gtk_clist_set_selection_mode (list, GTK_SELECTION_MULTIPLE);

	for (i=0; i!=row_qty; ++i)
	{
		gint row = gtk_clist_insert (list, -1, &rows[i].text);

		gulong* plong = g_new0 (gulong, 1);
		*plong = rows[i].flag;
		gtk_clist_set_row_data_full (list, row, plong, g_free);

		if (rows[i].flag & flags)
			gtk_clist_select_row ( list, row, -1 );
	}

	gtk_signal_connect ( GTK_OBJECT(list), "select-row",
		GTK_SIGNAL_FUNC(header_selection_changed), NULL );
	gtk_signal_connect ( GTK_OBJECT(list), "unselect-row",
		GTK_SIGNAL_FUNC(header_selection_changed), NULL );

	label = gtk_label_new (_("Select the fields you would like to see at the top of each message that is displayed in the message pane.") );
	gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);

	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET(list), FALSE, FALSE, GNOME_PAD_SMALL);

	return vbox;
}

static void
clear_acache_clicked_cb (GtkButton *button, gpointer data)
{
	const int i = acache_expire_all ();
	if (i != 0)
	{
		gchar* buf = g_strdup_printf ("%d files deleted.", i);
		gnome_ok_dialog (buf);
		g_free (buf);
	}
}

/* Build and return the 'Cache' page */
static GtkWidget*
prefs_cache_page (void)
{
	const int cache_max = gnome_config_get_int_with_default("/Pan/Cache/MaxSize=1242880", NULL);
	const gboolean flush_on_exit = gnome_config_get_bool("/Pan/Cache/FlushOnExit=true");
	GtkWidget *label;
	GtkWidget *vbox = gtk_vbox_new (FALSE, GNOME_PAD);
	GtkWidget *table;
	GtkWidget *hbox;
	GtkWidget *button;
	GtkWidget *tb;
	char buf[16];

	label = gtk_label_new (_("The cache is used to keep local copies of articles\nand thus reduce the time connected to the network."));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), GNOME_PAD_SMALL);

	table = gtk_table_new (3, 3, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);
	label = gtk_label_new (_("Disk Cache:"));
	gtk_widget_show (label);
	gtk_table_attach_defaults (GTK_TABLE(table), label, 0, 1, 0, 1);

	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	prefs_window->cache_kbytes_entry = gtk_entry_new();
	gtk_widget_show (prefs_window->cache_kbytes_entry);
	sprintf (buf, "%d", cache_max);
	gtk_entry_set_text (GTK_ENTRY(prefs_window->cache_kbytes_entry), buf);
	gtk_signal_connect_object (GTK_OBJECT(prefs_window->cache_kbytes_entry),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));
	gtk_box_pack_start (GTK_BOX(hbox), prefs_window->cache_kbytes_entry, FALSE, FALSE, GNOME_PAD_SMALL);
	label = gtk_label_new (_("bytes"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_widget_show (hbox);
	gtk_table_attach_defaults (GTK_TABLE(table), hbox, 1, 2, 0, 1);

	button = gtk_button_new_with_label (_("Clear Disk Cache"));
	gtk_signal_connect (GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC (clear_acache_clicked_cb),
			    NULL);
	gtk_widget_show (button);
	gtk_table_attach_defaults (GTK_TABLE(table), button, 2, 3, 0, 1);
	gtk_widget_show (table);

	tb = prefs_window->flush_cache_on_exit_check =
		gtk_check_button_new_with_label (_("Flush Cache when exiting Pan"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(tb), flush_on_exit);
	gtk_signal_connect_object (GTK_OBJECT(tb), "toggled",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT (prefs_window->box));
	gtk_table_attach_defaults (GTK_TABLE(table), tb, 0, 3, 1, 2);


	gtk_box_pack_start (GTK_BOX(vbox), table, FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_widget_show (vbox);
	return vbox;
}

/* Build and return the 'SMTP Server' page */
static GtkWidget *
prefs_smtp_page (void)
{
	GtkWidget *table;
	GtkWidget *label;
	gchar *tmp;

	table = gtk_table_new (2, 2, FALSE);
        gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
        gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER (table), GNOME_PAD_SMALL);


	label = gtk_label_new (_("Outgoing Mail (SMTP) Server:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, 4, 0);

	prefs_window->smtp_address = gtk_entry_new ();
	if ((tmp = gnome_config_get_string ("/Pan/Mail/smtp_address"))) {
		gtk_entry_set_text (GTK_ENTRY (prefs_window->smtp_address), tmp);
		g_free (tmp);
	}
	gtk_table_attach (GTK_TABLE (table), prefs_window->smtp_address, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	gtk_signal_connect_object (GTK_OBJECT(prefs_window->smtp_address),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));

	label = gtk_label_new (_("SMTP Port:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
			  GTK_FILL, GTK_FILL, 4, 0);

	prefs_window->smtp_port = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY (prefs_window->smtp_port), "25");
	gtk_table_attach (GTK_TABLE (table), prefs_window->smtp_port, 1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 0);

	gtk_signal_connect_object (GTK_OBJECT(prefs_window->smtp_port),
				   "changed",
				   GTK_SIGNAL_FUNC(gnome_property_box_changed),
				   GTK_OBJECT(prefs_window->box));

	return table;
}

/*---[ prefs_server_page ]--------------------------------------------
 * This is the Server tab of the preferences.  It has a list of the
 * server's and buttons that allow users to make a new server entry,
 * edit a previous one, or delete one.  This doesn't actually let
 * them switch servers, that's done from the main menu of Pan.
 *--------------------------------------------------------------------*/
static GtkWidget *
prefs_server_page (void)
{
	GtkWidget *vbox, *hbox, *button;

	char *titles[2];
	titles[0] = _("Name");
	titles[1] = _("Address");

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox);

	prefs_window->server_clist = gtk_clist_new_with_titles (2, titles);
	gtk_clist_column_titles_passive (GTK_CLIST(prefs_window->server_clist));
	gtk_clist_set_selection_mode (GTK_CLIST(prefs_window->server_clist),
				      GTK_SELECTION_SINGLE);

	gtk_clist_set_column_width (GTK_CLIST(prefs_window->server_clist), 0, 100);
	gtk_clist_set_column_auto_resize (GTK_CLIST(prefs_window->server_clist), 1, TRUE);

	gtk_widget_set_usize (prefs_window->server_clist, 280, 200);
	gtk_box_pack_start (GTK_BOX(hbox), prefs_window->server_clist, TRUE,
			    TRUE, 5);
	gtk_signal_connect (GTK_OBJECT(prefs_window->server_clist),
			    "button_press_event",
			    GTK_SIGNAL_FUNC(server_clist_button_press), NULL);

	prefs_create_clist ();

	/* one button vbox */
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), vbox, FALSE, FALSE, 5);

	/* new server button */
	button = gtk_button_new_with_label (_("New..."));
	gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, GNOME_PAD);
	gtk_signal_connect (GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC (new_server_cb), NULL);

	/* edit server button */
	button = gtk_button_new_with_label (_("Edit..."));
	gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, GNOME_PAD);
	gtk_signal_connect (GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC (edit_server_cb), NULL);

	/* delete server button */
	button = gtk_button_new_with_label(_("Delete"));
	gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, GNOME_PAD);
	gtk_signal_connect (GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC (delete_server_cb), NULL);

	return hbox;
}

void
prefs_spawn_to_news_connections (void)
{
	prefs_spawn ();

	pan_lock ();
	gtk_clist_select_row (GTK_CLIST(sections_clist), nntp_connections_row, 0);
	pan_unlock ();
}

void
prefs_spawn_to_news_servers (void)
{
	prefs_spawn ();

	pan_lock ();
	gtk_clist_select_row (GTK_CLIST(sections_clist), nntp_clist_row, 0);
	pan_unlock ();
}

/**
 * prefs_spawn:
 *
 * Build and draw the main Preferences dialog.  Its a GnomePropertyBox
 * with a notebook in it.
 **/
void
prefs_spawn (void)
{
	GtkWidget *hbox;

	/* There can be only one! (properties window) */
	if (prefs_window) {
		gdk_window_raise (GTK_WIDGET(prefs_window->box)->window);
		return;
	}

	prefs_window = g_new0 (PrefsWindow, 1);
	prefs_window->box = GNOME_PROPERTY_BOX (gnome_property_box_new());

	gtk_window_set_title (GTK_WINDOW (prefs_window->box), _("Pan: Preferences"));

	hbox = gtk_hbox_new (FALSE, GNOME_PAD);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);

	sections_clist = gtk_clist_new (1);
	gtk_clist_set_selection_mode (GTK_CLIST (sections_clist),
				      GTK_SELECTION_BROWSE);
	gtk_signal_connect (GTK_OBJECT (sections_clist), "select_row",
			    GTK_SIGNAL_FUNC (prefs_select_row_cb), NULL);


	/* add all the pages... */
	prefs_add_section_page (_("Cache"), prefs_cache_page());
	prefs_add_section_page (_("Directories"), prefs_directories_page());
	prefs_add_section_page (_("Filling/Wrapping Text"), prefs_fill_page());
	prefs_add_section_page (_("Fonts and Colors"), prefs_display_page());
	prefs_add_section_page (_("General"), prefs_general_page());
	prefs_add_section_page (_("Headers"), prefs_message_page());
	prefs_add_section_page (_("Mail Server"), prefs_smtp_page());
	nntp_clist_row = prefs_add_section_page (_("News Servers"),
	                                         prefs_server_page());
	nntp_connections_row =prefs_add_section_page (_("Online Settings"),
			                              prefs_connections_page());
	prefs_add_section_page (_("User"), prefs_userinfo_page());

	gtk_notebook_set_show_tabs (GTK_NOTEBOOK(prefs_window->box->notebook),
				    FALSE);
	gtk_notebook_set_show_border (GTK_NOTEBOOK(prefs_window->box->notebook),
				      FALSE);

	gtk_signal_connect (GTK_OBJECT (prefs_window->box), "apply",
			    GTK_SIGNAL_FUNC (prefs_apply_cb), NULL);
	gtk_signal_connect (GTK_OBJECT (prefs_window->box), "destroy",
			    GTK_SIGNAL_FUNC (prefs_cancel_cb), NULL);


	gtk_widget_ref (GNOME_PROPERTY_BOX (prefs_window->box)->notebook);
	
	gtk_container_remove (GTK_CONTAINER (GNOME_DIALOG (prefs_window->box)->vbox),
			      GNOME_PROPERTY_BOX (prefs_window->box)->notebook);

	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG(prefs_window->box)->vbox),
			    hbox, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX (hbox),
			    sections_clist,
			    TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX (hbox),
			    GNOME_PROPERTY_BOX (prefs_window->box)->notebook,
			    TRUE, TRUE, 0);

	gtk_widget_unref (GNOME_PROPERTY_BOX (prefs_window->box)->notebook);
			    
	gtk_widget_set_usize (GTK_WIDGET (sections_clist), 150, 0);

	gui_popup_draw (GTK_WIDGET (prefs_window->box), Pan.window);
}
