/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-config.h>
#include <libgnomeui/libgnomeui.h>

#include "acache.h"
#include "articlelist.h"
#include "gui.h"
#include "gui-headers.h"
#include "globals.h"
#include "grouplist.h"
#include "message-window.h"
#include "pan-object.h"
#include "prefs.h"
#include "queue.h"
#include "task-body.h"
#include "text.h"
#include "util.h"

static GdkColor headers_col = {0, 0xbbbb, 0, 0};
static GtkWidget * header_table = NULL;

static void ensure_header_col_inited (void);
static void realize_text_cb (GtkWidget *text, gpointer data);
static void text_button_press_cb (GtkWidget *widget, GdkEventButton *bevent);
static void text_key_press_cb (GtkWidget *widget, GdkEventKey *event);
static void text_page_down_or_next_article (GtkText *text);
static void show_all_headers_cb (GtkCheckMenuItem* item, gpointer data);

static void
rot13_cb (GtkCheckMenuItem* item, gpointer data)
{
	text_set_rot13 (item->active);
}

/***
****
****  POPUP MENUS
****
***/

static GnomeUIInfo respond_menu [] =
{
        {
		GNOME_APP_UI_ITEM,
		N_("Followup to newsgroup"),
		N_("Post a reply to the message on the news server."),
		message_followup_window, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL_RPL,
		0, 0, NULL
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Reply by E-mail"),
		N_("Create a mail reply to the sender."),
		message_reply_window, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL_RPL,
		0, 0, NULL
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Followup to newsgroup and Reply by E-mail"),
		N_("Send a reply both to the author in mail, and to the news server."),
		message_followup_reply_window, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL_RPL,
		0, 0, NULL
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Forward article by E-mail"),
		N_("Forward article by E-mail"),
		message_forward_window
	},
	GNOMEUIINFO_END
};

static GnomeUIInfo search_menu[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("Search for user on deja"),
		N_("Search for other postings by this use on deja.com"),
		search_deja,
	},
	GNOMEUIINFO_END
};

static GnomeUIInfo filter_menu[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("Watch Thread"),
		N_("Watch Thread"),
		articlelist_selected_thread_watch
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Ignore Thread"),
		N_("Ignore Thread"),
		articlelist_selected_thread_ignore
	},
	GNOMEUIINFO_END
};


static GnomeUIInfo text_menu_popup[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("Open in New Window"),
		N_("Open this message for reading."),
		message_read_window
	},
	GNOMEUIINFO_SEPARATOR,
        {
		GNOME_APP_UI_SUBTREE,
		N_("Respond"),
		N_("Respond"),
		&respond_menu, NULL,
		NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK,
		'\0', 0, NULL
	},
	{
		GNOME_APP_UI_SUBTREE,
		N_("Search Offsite"),
		N_("Search Offsite"),
		&search_menu, NULL,
		NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK,
		'\0', 0, NULL
	},
	{
		GNOME_APP_UI_SUBTREE,
		N_("Filter"),
		N_("Filter"),
		&filter_menu, NULL,
		NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK,
		'\0', 0, NULL
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_TOGGLEITEM,
		N_("Show All Headers in Body"),
		N_("Show All Headers in Body"),
		show_all_headers_cb
	},
	{
		GNOME_APP_UI_TOGGLEITEM,
		N_("Rot13 the Body"),
		N_("Rot13 the Body"),
		rot13_cb
	},
#if 0
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Select All"),
		N_("Select all of the message body"),
		text_select_all,
		NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL,
		'A', GDK_CONTROL_MASK, NULL
	},
#endif
	GNOMEUIINFO_END
};

static gchar welcome[] =
{
"Pan " VERSION "\n"
"Copyright (C) 1999, 2000 \n"
"Pan Development Team (pan@superpimp.org) \n"
"\n"
"  This is beta software.  If you find a bug, please report it.\n"
"\n"
"  http://www.superpimp.org/contact.html   -  Request a Feature\n"
"  http://www.superpimp.org/bugreport.html -  Report a Bug\n"
"  http://www.superpimp.org/download.html  -  Upgrade Pan\n"
"\n"
"This program is free software; you can redistribute it \n"
"and/or modify it under the terms of the GNU General Public \n"
"License as published by the Free Software Foundation; \n"
"either version 2 of the License, or (at your option) any \n"
"later version. \n"
"\n"
"This program is distributed in the hope that it will be \n"
"useful, but WITHOUT ANY WARRANTY; without even the implied \n"
"warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR \n"
"PURPOSE.  See the GNU General Public License for more \n"
"details. \n"
"\n"
"You should have received a copy of the GNU General Public \n"
"License along with this program; if not, write to the \n"
"Free Software Foundation, Inc., 59 Temple Place, \n"
"Suite 330, Boston, MA 02111-1307, USA"
};

static GtkWidget * text_rightclick_popup = NULL;
PanCallback * text_show_all_headers_changed = NULL;
PanCallback * current_article_changed = NULL;
gboolean rot13_enabled = FALSE;
gboolean show_all_headers_in_body = FALSE;

static Article * current_article = NULL;
GdkColor text_fg_color;
GdkColor text_bg_color;
GdkColor text_quoted_color;


static void
show_all_headers_cb (GtkCheckMenuItem * item, gpointer data)
{
	text_set_show_all_headers_in_body (item->active);
}

void
text_select_all (void)
{
	pan_lock();
	gtk_editable_select_region (GTK_EDITABLE(Pan.text), 0, -1);
	pan_unlock();
}

void
text_refresh (void)
{
	if (current_article != NULL)
	{
		text_update (current_article);
	}
	else
	{
		char* contents = gtk_editable_get_chars (
			GTK_EDITABLE(Pan.text),0,-1);
		update_body_pane (Pan.text, contents);
		g_free (contents);
	}
}

void
text_set_rot13 (gboolean on)
{
	if (on != rot13_enabled)
	{
		rot13_enabled = on;

		if (current_article != NULL)
			text_update (current_article);
	}
}

void
text_set_show_all_headers_in_body (gboolean on)
{
	if (on != show_all_headers_in_body)
	{
		show_all_headers_in_body = on;

		if (current_article != NULL)
			text_update (current_article);

		/* update our menu item */
		if (1) {
			gboolean show_on = on;
			GtkCheckMenuItem * menu = GTK_CHECK_MENU_ITEM(text_menu_popup[6].widget);
			if (show_on != menu->active)
				gtk_check_menu_item_set_active (menu, show_on);
		}

		/* notify listeners */
		pan_callback_call (text_show_all_headers_changed,
		                   NULL,
		                   GINT_TO_POINTER(show_all_headers_in_body));
	}
}

void
text_read_more (void)
{
	text_page_down_or_next_article (GTK_TEXT(Pan.text));
}

/*
 * Scroll down one page, or if we're at the bottom,
 * read the next article
 */
static void
text_page_down_or_next_article (GtkText *text)
{
        gfloat upper;
	gfloat old_val;

	g_return_if_fail (text != NULL);

	/* calculate a page down */
	old_val = text->vadj->value;
	text->vadj->value += text->vadj->page_increment;
	upper = text->vadj->upper - text->vadj->page_size;
        text->vadj->value = MIN (text->vadj->value, upper);
        text->vadj->value = MAX (text->vadj->value, 0.0);

	/* either go down a page or, if we're at the end, read the next article */
	if (text->vadj->value != old_val)
		gtk_signal_emit_by_name (GTK_OBJECT (text->vadj), "value_changed");
	else
		articlelist_view_next ();
}

/* Called when the widget is first drawn, puts our welcome text into view */
static void
realize_text_cb (GtkWidget *text, gpointer data)
{
	GtkAdjustment* hadj = GTK_ADJUSTMENT(GTK_TEXT(text)->hadj);
	GtkAdjustment* vadj = GTK_ADJUSTMENT(GTK_TEXT(text)->vadj);

	gui_headers_set_default (header_table, NULL);
	gtk_editable_delete_text (GTK_EDITABLE(text), 0, -1);
	gtk_text_insert (
		GTK_TEXT(text), NULL, &text_fg_color, NULL, welcome, -1);
	gtk_text_set_point (GTK_TEXT(text), 0);
	gtk_adjustment_set_value (GTK_ADJUSTMENT(vadj), 0);
	gtk_text_set_adjustments (GTK_TEXT(text), hadj, vadj);

	gtk_signal_disconnect_by_func (
		GTK_OBJECT(text), GTK_SIGNAL_FUNC(realize_text_cb), NULL);
}

/* Process a mouse button events inside the text widget */
static void
text_button_press_cb (GtkWidget *widget, GdkEventButton *bevent)
{
	gtk_widget_grab_focus (GTK_WIDGET (Pan.text));

	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (
			GTK_OBJECT(widget), "button_press_event");
	
	if (bevent->button == 3)
	{
		/* show the menu */
		gnome_popup_menu_do_popup (text_rightclick_popup, NULL,
					   NULL, bevent, NULL);
	}
}

/* Process a keyboard input event in the text pane */
static void
text_key_press_cb (GtkWidget *widget, GdkEventKey *event)
{
#if 0
	switch (event->keyval)
	{
		case GDK_KP_Space:
		case GDK_space:
			text_page_down_or_next_article (GTK_TEXT (Pan.text));
			break;
		default: 
			break;
	}
#endif
}

/**
***  CURRENT ARTICLE
**/

Article*
get_current_article (void)
{
	return current_article;
}

void
set_current_article (Article * a)
{
	Article * old_article;

	if (a!=NULL)
		group_ref_articles (a->group, NULL);

	old_article = current_article;
	current_article = a;
	pan_callback_call (current_article_changed, old_article, current_article);

	if (old_article != NULL)
		group_unref_articles (old_article->group, NULL);
}

static int
group_articles_removed_cb (gpointer call_obj,
                           gpointer call_arg,
                           gpointer client_data)
{
	Group * group = GROUP(call_obj);
	GPtrArray * removed = (GPtrArray*) call_arg;

	/* sanity checks */
	g_return_val_if_fail (group!=NULL, 0);
	g_return_val_if_fail (removed!=NULL, 0);
	g_return_val_if_fail (removed->len>0, 0);

	/* unset the current article if we need to */
	if (current_article!=NULL) {
		guint i;
		gboolean current_removed = FALSE;
		for (i=0; !current_removed && i!=removed->len; ++i)
		if (ARTICLE(g_ptr_array_index(removed,i)) == current_article)
			current_removed = TRUE;
		if (current_removed) {
			set_current_article (NULL);
			update_body_pane (Pan.text, NULL);
			gui_headers_set_default (header_table, NULL);
		}
	}

	return 0;
}

/**
***
**/

GtkWidget *
text_create (void)
{
	GtkStyle * style;
	GtkWidget * text_box;
	GtkWidget *scrolled_window;

	/* setup */
	ensure_header_col_inited();

	/* create the text window, set its properties, connect signals */
	Pan.text = gtk_text_new (NULL, NULL);
	gtk_text_set_editable (GTK_TEXT(Pan.text), FALSE);
	widget_set_font (GTK_WIDGET (Pan.text), message_body_font);
	style = gtk_widget_get_style(GTK_WIDGET(Pan.text));
	style->text[0] = text_fg_color;
	style->base[0] = text_bg_color;
	text_show_all_headers_changed = pan_callback_new ();
	current_article_changed = pan_callback_new ();
	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_container_set_border_width (GTK_CONTAINER(scrolled_window), GNOME_PAD_SMALL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "realize",
			    GTK_SIGNAL_FUNC(realize_text_cb), NULL);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "button_press_event",
			    GTK_SIGNAL_FUNC(text_button_press_cb), NULL);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "key_press_event",
			    GTK_SIGNAL_FUNC(text_key_press_cb), NULL);
	gtk_container_add (GTK_CONTAINER(scrolled_window), Pan.text);
	text_rightclick_popup = gnome_popup_menu_new (text_menu_popup);

	/* text_box holds the header info and the scrolled text window */
	text_box = gtk_vbox_new (FALSE, 0);
	header_table = gtk_table_new ( 0, 0, FALSE);
	gtk_box_pack_start (GTK_BOX(text_box), header_table, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(text_box), scrolled_window, TRUE, TRUE, 0);

	/* watch for dying articles */
	pan_callback_add (group_articles_removed,
	                  group_articles_removed_cb,
	                  NULL);

	return text_box;
}

/* Do ROT13 shifting on a message body */
static void
rot13 (gchar* body)
{
	const gchar * a = "abcdefghijklmnopqrstuvwxyz";
	register gchar *pch;

	for (pch=body; *pch!='\0'; ++pch)
	{
		if (isalpha((int)*pch))
		{
			const gboolean uc = isupper((int)*pch);
			*pch = tolower(*pch);
			*pch = a[((*pch-'a')+13)%26];
			if (uc)
				*pch = toupper(*pch);
		}
	}
}

static void
ensure_header_col_inited (void)
{
	static gboolean headers_col_inited = FALSE;

	if (!headers_col_inited)
	{
		pan_lock ();
		if (gdk_color_alloc (cmap, &headers_col))
			headers_col_inited = TRUE;
		else
			g_error (_("couldn't allocate \"headers\" color"));
		pan_unlock ();
	}
}

/**
***
**/

void
text_update (Article *article)
{
	gchar *body;

	pan_object_ref (PAN_OBJECT(article));

        /* sanity checks */
        g_return_if_fail (article!=NULL);
        g_return_if_fail (article->message_id!=NULL);
        g_return_if_fail (*article->message_id!='\0');
	
        /* load article body */
	body = article_get_body (article);
	if (body != NULL)
	{
		/* update the current article */
		set_current_article (article);

		/* rot13 */
		if (rot13_enabled)
			rot13 (body);

		/* fill */
		if (fill_body_on_show)
			replace_gstr (&body, fill_body(body));

		/* add headers */
		if (show_all_headers_in_body) {
			gchar* headers = acache_load_header (article->message_id);
			if (headers != NULL) { /* add headers to body */
				replace_gstr (&body, g_strconcat(headers,"\n\n", body, 0));
				replace_gstr (&headers, NULL);
			}
		}

		/* update main text window */
		update_body_pane (Pan.text, body);
		gui_page_set (MESSAGE_PAGE, Pan.text);
		gui_headers_set_default (header_table, article);

		/* cleanup */
		g_free (body);
	}

	pan_object_unref (PAN_OBJECT(article));
}

void
text_set_from_article (Article *article,
		       MessageWindowType type,
		       gboolean needs_window)
{
	/* sanity checks */
	g_return_if_fail (article!=NULL);

	queue_add (TASK(task_body_new(article, needs_window, type)));
}
